//##############################################################################
// siphon
//
// common.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#ifndef NDEBUG

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <SDL.h>

#include "common.h"

//##############################################################################
// Linked List with void * value
//##############################################################################
typedef struct Node
{
	void * name;
	char * value;
	Uint32 timestamp;
	size_t bytes;
	struct Node * next;
} Node;

static Node * get_node(Node * list, void * name)
{
	while(list != NULL)
	{
		if(list->name == name)
			return list;
		list = list->next;
	}
	return NULL;
}

static void append_node(Node * list, Node * node)
{
	assert(list != NULL);
	while(list->next != NULL)
	{
		list = list->next;
	}
	list->next = node;
}

static void delete_node(Node * node)
{
	assert(node != NULL);
	free(node->value);
	free(node);
}

static Node * create_node(void * name, const char * value, size_t bytes)
{
	Node * node = (Node*)malloc(sizeof(Node));
	node->name = name;
	node->value = strdup(value);
	node->timestamp = SDL_GetTicks();
	node->bytes = bytes;
	node->next = NULL;
	return node;
}

//##############################################################################
// Static variables
//##############################################################################
static unsigned int m_allocations = 0;
static Node * m_named_allocations = NULL;
static size_t m_bytes = 0;

//##############################################################################
// Exported functions
//##############################################################################
void * siphon_malloc_l(size_t bytes, const char * line)
{
	assert(bytes > 0);

	void * ptr = malloc(bytes);

	Node * node = create_node(ptr, line, bytes);
	if(m_named_allocations == NULL)
		m_named_allocations = node;
	else
		append_node(m_named_allocations, node);

	m_allocations++;
	m_bytes += bytes;
	return ptr;
}

void siphon_free(void * ptr)
{
	assert(ptr != NULL);

	Node * prev = NULL;
	Node * node = m_named_allocations;
	size_t bytes = 0;

	if(ptr == (void*)0xFFFFFFFF)
	{
		fprintf(stderr, "Trying to free already freed memory.\n");
		fprintf(stderr, "Use gdb and 'backtrace' to track down the cause.\n");
		fprintf(stderr, "Or use PRELOAD with libsigsegv.so.\n");
		abort();
	}

	while(node)
	{
		if(ptr == node->name)
		{
			if(prev == NULL)
			{
				m_named_allocations = node->next;
				bytes = node->bytes;
				delete_node(node);
			}
			else
			{
				prev->next = node->next;
				bytes = node->bytes;
				delete_node(node);
			}
			break;
		}
		prev = node;
		node = node->next;
	}

	m_bytes -= bytes;
	m_allocations--;

	// Write 0xFF's to the memory to increase the likelyhood of crashes if
	// freed memory is accessed.
	memset(ptr, 0xFF, bytes);

	free(ptr);
}

void * siphon_realloc(void * ptr, size_t size)
{
	assert(size > 0);

	if(ptr == NULL)
		return siphon_malloc(size);

	Node * node = get_node(m_named_allocations, ptr);
	assert(node != NULL);
	node->name = realloc(ptr, size);

	m_bytes -= node->bytes;
	node->bytes = size;
	m_bytes += size;
	node->timestamp = SDL_GetTicks();

	return node->name;
}

char * siphon_strdup_l(const char * str, const char * line)
{
	size_t bytes = strlen(str) + 1;
	char * cpy = strdup(str);

	Node * node = create_node((void*)cpy, line, bytes);
	if(m_named_allocations == NULL)
		m_named_allocations = node;
	else
		append_node(m_named_allocations, node);

	m_bytes += bytes;
	m_allocations++;
	return cpy;
}

void siphon_print_allocations()
{
	Node * node = m_named_allocations;

	while(node)
	{
		fprintf(stdout, "%u ms, %u bytes, %s\n",
			node->timestamp, node->bytes, node->value);
		node = node->next;
	}

	fprintf(stdout, "Allocations: %d\n", m_allocations);
	fprintf(stdout, "Total Bytes: %u\n", m_bytes);
}

#endif
