//##############################################################################
// siphon
//
// datastructures.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <string.h>
#include <assert.h>

#include "datastructures.h"
#include "common.h"

//##############################################################################
// Static functions (SDict)
//##############################################################################
static SDict * sdict_create_node(const char * key, void * value)
{
	SDict * node = (SDict*)siphon_malloc(sizeof(SDict));
	node->key = siphon_strdup(key);
	node->value = value;
	node->next = NULL;
	return node;
}

static void sdict_free_node(SDict * node)
{
	assert(node != NULL);
	siphon_free(node->key);
	siphon_free(node);
}

//##############################################################################
// Exported functions (SDict)
//##############################################################################
SDict * sdict_add(SDict * dict, const char * key, void * value)
{
	SDict * node = NULL;

	assert(key != NULL);

	node = sdict_create_node(key, value);
	node->next = dict;
	return node;
}

SDict * sdict_find(SDict * dict, const char * key)
{
	assert(key != NULL);
	
	while(dict != NULL)
	{
		if(strcmp(dict->key, key) == 0)
			return dict;
		dict = dict->next;
	}
	return NULL;
}

SDict * sdict_free(SDict * dict)
{
	while(dict != NULL)
	{
		SDict * next = dict->next;
		sdict_free_node(dict);
		dict = next;
	}
	return NULL;
}

SDict * sdict_free_full(SDict * dict, void(*free_func)(void*))
{
	assert(free_func != NULL);
	
	while(dict != NULL)
	{
		SDict * next = dict->next;
		free_func(dict->value);
		sdict_free_node(dict);
		dict = next;
	}
	return NULL;
}

//##############################################################################
// Static functions (SList)
//##############################################################################
static SList * slist_create_node(void * data)
{
	SList * node = (SList*)siphon_malloc(sizeof(SList));
	node->data = data;
	node->next = NULL;
	return node;
}

static void slist_free_node(SList * node)
{
	assert(node != NULL);
	siphon_free(node);
}

//##############################################################################
// Exported functions (SList)
//##############################################################################
SList * slist_add(SList * list, void * data)
{
	SList * node = NULL;

	assert(data != NULL);

	node = slist_create_node(data);
	node->next = list;
	return node;
}

SList * slist_remove(SList * list, void * data)
{
	SList * iter = list;
	SList * prev = NULL;
	
	while(iter != NULL)
	{
		if(iter->data == data)
		{
			if(prev == NULL)
			{
				list = iter->next;
				slist_free_node(iter);
			}
			else
			{
				prev->next = iter->next;
				slist_free_node(iter);
			}
			break;
		}
		prev = iter;
		iter = iter->next;
	}
	return list;
}

SList * slist_find(SList * list, slist_compare_func cb, void * userdata)
{
	assert(cb != NULL);

	while(list != NULL)
	{
		if(cb(list->data, userdata) == 0)
			return list;
		list = list->next;
	}
	return NULL;
}

SList * slist_free(SList * list)
{
	while(list != NULL)
	{
		SList * next = list->next;
		slist_free_node(list);
		list = next;
	}
	return NULL;
}

SList * slist_free_full(SList * list, void(*free_func)(void*))
{
	assert(free_func != NULL);
	
	while(list != NULL)
	{
		SList * next = list->next;
		free_func(list->data);
		slist_free_node(list);
		list = next;
	}
	return NULL;
}
