//##############################################################################
// siphon
//
// font.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <string.h>
#include <assert.h>

#include "font.h"
#include "resources.h"
#include "common.h"

//##############################################################################
// Static functions
//##############################################################################
static int get_pixel(SDL_Surface * surface, unsigned int x, unsigned int y)
{
	if(x > (unsigned int)surface->w || y > (unsigned int)surface->h)
	{
		fprintf(stderr, "Tried accessing a pixel out of surface bounds\n");
		exit(EXIT_FAILURE);
	}

	switch(surface->format->BytesPerPixel)
	{
		case 1:
			return ((uint8_t*)surface->pixels)[x + y * surface->w];
		case 2:
			return ((uint16_t*)surface->pixels)[x + y * surface->w];
		case 4:
			return ((uint32_t*)surface->pixels)[x + y * surface->w];
		default:
			fprintf(stderr, "Could not handle pixel format\n");
			exit(EXIT_FAILURE);
	}
	return -1;
}

//##############################################################################
// Exported functions
//##############################################################################
Font * font_create(const char * name, const char * alphabet)
{
	Font * font = NULL;
	SDL_Surface * surface = NULL;
	int characters = 0;
	int i = 0;
	int x = 0;
	GLint tex_width = 0;
	GLint tex_height = 0;
	int magic_color = 0;
	int elements = 0;

	assert(name != NULL);
	assert(alphabet != NULL);

	font = (Font*)siphon_malloc(sizeof(Font));
	font->highestindex = 0;
	font->height = 0;
	font->bottom = 0.0f;
	font->lefts = NULL;
	font->rights = NULL;
	font->tex = NULL;

	surface = resources_get_image(name);

	characters = strlen(alphabet);

	// Allocate exactly enough bytes to fit the font
	for(i = 0; i < characters; i++)
	{
		if(alphabet[i] > font->highestindex)
			font->highestindex = alphabet[i];
	}


	elements = font->highestindex + 1;
	font->lefts = (float*)siphon_malloc(elements * sizeof(float));
	font->rights = (float*)siphon_malloc(elements * sizeof(float));
	font->widths = (int*)siphon_malloc(elements * sizeof(int));

	// Empty arrays
	for(i = 0; i < font->highestindex + 1; i++)
	{
		font->widths[i] = 0;
		font->lefts[i] = 0;
		font->rights[i] = 0;
	}

	// Load gl texture and retrieve width and height
	font->tex = resources_get_gl_image(name);
	glBindTexture(GL_TEXTURE_2D, *(font->tex));
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &tex_width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &tex_height);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Set height as texture coordinate
	font->height = surface->h;
	font->bottom = surface->h / (float)tex_height;

	// Read font
	magic_color = get_pixel(surface, 0, 0);
	for(i = 0; alphabet[i]; i++)
	{
		int startx, endx;
		int index = alphabet[i];

		// read magic_color colored bar before character
		for(; get_pixel(surface, x, 0) == magic_color && x < surface->w; x++);
		startx = x;

		// read until the next magic_color colored bar
		for(; get_pixel(surface, x, 0) != magic_color && x < surface->w; x++);
		endx = x;

		font->widths[index] = endx - startx;
		font->lefts[index] = startx / (float)tex_width;
		font->rights[index] = endx / (float)tex_width;
	}

	return font;
}

void font_destroy(Font * font)
{
	assert(font != NULL);
	siphon_free(font->lefts);
	siphon_free(font->rights);
	siphon_free(font->widths);
	siphon_free(font);
}

///TODO: This should be revised
void font_draw_dialog(Font * font, Vector pos, Vector size, const char * text)
{
	int lines = size.y / font->height;
	int i = 0;
	char * next = (char*)text;

	glBindTexture(GL_TEXTURE_2D, *(font->tex));
	glBegin(GL_QUADS);
	for(i = 0; i < lines; i++)
	{
		int current_width = 0;
		char * walker = next;
		char * current = next;
		int top = pos.y + i * font->height;
		int left = pos.x;

		if(next == NULL)
			break;

		while(*walker && current_width <= size.x)
		{
			current_width += font->widths[(int)*walker] + 1;
			walker++;
		}

		if(current_width > size.x)
		{
			walker--;
			while(*walker != ' ' && walker >= text)
			{
				walker--;
			}
			if((walker-1) >= text && *(walker-1) == ' ')
			{
				next = walker;
			}
			else if(*walker == ' ')
			{
				next = walker + 1;
			}
			else
			{
				break;
			}
		}
		else
		{
			next = NULL;
		}

		while(current != walker)
		{
			int index = (int)*current;

			if(index < 0 || index > font->highestindex ||
				font->widths[index] == 0)
			{
				current++;
				continue;
			}

			glTexCoord2f(font->lefts[index], 0.0f);
			glVertex2f(left, top);
			glTexCoord2f(font->rights[index], 0.0f);
			glVertex2f(left + font->widths[index], top);
			glTexCoord2f(font->rights[index], font->bottom);
			glVertex2f(left + font->widths[index], top + font->height);
			glTexCoord2f(font->lefts[index], font->bottom);
			glVertex2f(left, top + font->height);

			left += font->widths[index] + 1;
			current++;
		}
	}
	glEnd();
	glBindTexture(GL_TEXTURE_2D, 0);
}

void font_draw_text(Font * font, Vector pos, const char * text)
{
	int i = 0;

	assert(font != NULL);
	assert(text != NULL);

	glBindTexture(GL_TEXTURE_2D, *(font->tex));
	glBegin(GL_QUADS);
	while(*text)
	{
		if(*text < 0 || *text > font->highestindex ||
			font->widths[(int)*text] == 0)
		{
			text++;
			continue;
		}
		i = (int)(*text);

		glTexCoord2f(font->lefts[i], 0.0f);
		glVertex2f(pos.x, pos.y);

		glTexCoord2f(font->rights[i], 0.0f);
		glVertex2f(pos.x + font->widths[i], pos.y);

		glTexCoord2f(font->rights[i], font->bottom);
		glVertex2f(pos.x + font->widths[i], pos.y + font->height);

		glTexCoord2f(font->lefts[i], font->bottom);
		glVertex2f(pos.x, pos.y + font->height);

		pos.x += font->widths[i] + 1;
		text++;
	}
	glEnd();
	glBindTexture(GL_TEXTURE_2D, 0);
}

Vector font_measure_text(Font * font, const char * text)
{
	Vector size = { 0, 0 };

	assert(font != NULL);
	assert(text != NULL);

	size.y = font->height;

	while(*text)
	{
		if(*text >= 0 && *text <= font->highestindex)
			size.x += font->widths[(int)*text] + 1;
		text++;
	}

	return size;
}
