//##############################################################################
// siphon
//
// gameobject.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <assert.h>
#include <stdlib.h>

#include "gameobject.h"

//##############################################################################
// Exported functions
//##############################################################################
GameObjectNode * gameobject_create_node(GameObject * go)
{
	GameObjectNode * node = (GameObjectNode*)
		siphon_malloc(sizeof(GameObjectNode));
	node->go = go;
	node->next = NULL;
	return node;
}

void gameobject_append_node(GameObjectNode * list, GameObjectNode * node)
{
	assert(list != NULL);
	assert(node != NULL);

	while(list->next != NULL)
	{
		list = list->next;
	}
	list->next = node;
}

void gameobject_destroy(GameObjectNode * node)
{
	while(node)
	{
		GameObjectNode * current = node;
		node = node->next;

		if(current->go->history != NULL)
			siphon_free(current->go->history);
		if(current->go->events != NULL)
			siphon_free(current->go->events);
		assert(current->go->destroy != NULL);
		current->go->destroy(current->go);

		siphon_free(current);
	}
}

bool gameobject_collides(const GameObject * go, CollisionType type)
{
	Vector tl, br;
	int x = TOINT(go->pos.x);
	int y = TOINT(go->pos.y);

	assert(go != NULL);
	assert(go->map != NULL);

	switch(type)
	{
		case COLLISION_DOWN:
			y += 1;
			break;
		case COLLISION_LEFT:
			if(x <= 0)
				return FALSE;
			x -= 1;
			break;
		case COLLISION_RIGHT:
			if(x + go->size.x >= (int)(go->map->w * go->map->tilesize))
				return FALSE;
			x += 1;
			break;
		case COLLISION_UP:
			y -= 1;
			break;
	}

	br.x = (x + go->size.x - 1) >> go->map->tileshift;
	br.y = (y + go->size.y - 1) >> go->map->tileshift;

	for(tl.x = x >> go->map->tileshift; tl.x <= br.x; tl.x++)
	{
		for(tl.y = y >> go->map->tileshift; tl.y <= br.y; tl.y++)
		{
			int tilenum = map_get_tilenum(go->map, tl);
			if((tilenum == -1 || tilenum >= 4))
				return TRUE;
		}
	}

	return FALSE;
}

void gameobject_push_history(GameObject * go)
{
	unsigned int count;

	// Increase size of history array, also creates array if it doesn't exist
	if(go->history_count == go->history_size)
	{
		go->history_size = MAX(1024, go->history_size * 2);
		go->history = (GameObjectHistory*)siphon_realloc(
			go->history, sizeof(GameObjectHistory) * go->history_size);
	}

	count = go->history_count;

	go->history[count].pos = go->pos;
	go->history[count].vel = go->vel;

	go->history_count++;
}

void gameobject_pop_history(GameObject * go)
{
	unsigned int highestindex = go->history_count - 1;

	if(go->history_count == 0)
	{
		go->req_pos = go->pos;
		go->req_vel = go->vel;
		return;
	}

	go->req_pos = go->history[highestindex].pos;
	go->req_vel = go->history[highestindex].vel;

	go->history_count--;
}

void gameobject_push_event(GameObject * go, void * p1, void * p2,
	EventUndoFunc undo_func)
{
	if(go->events_is_popping)
		return;

	// Increase size of events array, also creates array if it doesn't exist
	if(go->events_count == go->events_size)
	{
		go->events_size = MAX(go->events_size * 2, 64);
		go->events = (Event*)siphon_realloc(go->events,
			sizeof(Event) * go->events_size);
	}

	go->events[go->events_count].tick = go->map->tickcount;
	go->events[go->events_count].p1 = p1;
	go->events[go->events_count].p2 = p2;
	go->events[go->events_count].undo_func = undo_func;

	go->events_count++;
}

void gameobject_pop_event(GameObject * go)
{
	Event * event = NULL;
	if(go->events_count == 0)
		return;

	event = go->events + go->events_count - 1;
	go->events_is_popping = TRUE;
	while(event->tick > go->map->tickcount)
	{
		event->undo_func(go, event);

		go->events_count--;

		if(go->events_count == 0)
			break;
		else
			event = event - 1;
	}
	go->events_is_popping = FALSE;
}
