//##############################################################################
// siphon
//
// bonus.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <assert.h>
#include <SDL.h>

#include "bonus.h"
#include "game.h"
#include "config.h"

//##############################################################################
// GameObject registration
//##############################################################################
CONSTRUCTOR(bonus_const)
{
	registry_register("Bonus", &bonus_create);
}

//##############################################################################
// Static functions
//##############################################################################
static void bonus_destroy(GameObject * vp_bonus)
{
	Bonus * bonus = (Bonus*)vp_bonus;
	assert(vp_bonus && vp_bonus->type & BONUS);
	sprite_destroy(bonus->sprite);
	siphon_free(bonus);
}

static void bonus_logic(GameObject * gameobject)
{
	FixedVector gravity;
	GameObject * go = gameobject;
	int touch_ground = gameobject_collides(go, COLLISION_DOWN);
	int touch_left = gameobject_collides(go, COLLISION_LEFT);
	int touch_right = gameobject_collides(go, COLLISION_RIGHT);
	int touch_top = gameobject_collides(go, COLLISION_UP);

	gameobject->req_pos = gameobject->pos;
	gameobject->req_vel = gameobject->vel;

	gravity.x = TOFIX(0) / 4;
	gravity.y = TOFIX(1) / 4;

	if(config.flip_gravity)
	{
		gravity.y = -gravity.y;
		gravity.x = -gravity.x;
	}

	if(touch_ground)
		go->req_vel.y = MIN(go->req_vel.y, 0);
	if(touch_top)
		go->req_vel.y = MAX(go->req_vel.y, 0);
	if(touch_left)
		go->req_vel.x = MAX(go->req_vel.x, 0);
	if(touch_right)
		go->req_vel.x = MIN(go->req_vel.x, 0);

	gameobject->req_vel.x += gravity.x;
	gameobject->req_vel.y += gravity.y;

	gameobject->req_vel.y = MAX(TOFIX(-7), gameobject->req_vel.y);
	gameobject->req_vel.y = MIN(TOFIX(7), gameobject->req_vel.y);

	gameobject->req_pos.x += gameobject->req_vel.x;
	gameobject->req_pos.y += gameobject->req_vel.y;
}

static void bonus_draw(GameObject * vp_bonus, int offset_x, int offset_y)
{
	Bonus * bonus = (Bonus*)vp_bonus;
	int b;
	assert(vp_bonus && vp_bonus->type & BONUS);

	b = (TOINT(vp_bonus->pos.y) - offset_y) * 2 + vp_bonus->size.y;

	if(config.flip_gravity)
	{
		glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
		glTranslatef(0.0f, -b, 0.0f);
	}

	sprite_draw(bonus->sprite, TOINT(vp_bonus->pos.x) - offset_x,
		TOINT(vp_bonus->pos.y) - offset_y);

	if(config.flip_gravity)
	{
		glTranslatef(0.0f, b, 0.0f);
		glRotatef(-180.0f, 1.0f, 0.0f, 0.0f);
	}
}

//##############################################################################
// Exported functions
//##############################################################################
GameObject * bonus_create(const CreateArgs * args)
{
	Bonus * bonus = (Bonus*)siphon_malloc(sizeof(Bonus));
	GameObject * gameobject = (GameObject*)bonus;
	const char * bonus_type = properties_get_value(args->properties, "type");
	Vector position = properties_get_vector(args->properties, "position");

	memset(bonus, 0, sizeof(Bonus));
	gameobject->map = args->map;
	gameobject->pos = vector_to_fixed(position);
	gameobject->req_pos = vector_to_fixed(position);
	gameobject->type = BONUS;
	gameobject->size.x = 4;
	gameobject->size.y = 4;
	gameobject->logic = &bonus_logic;
	gameobject->draw = &bonus_draw;
	gameobject->destroy = &bonus_destroy;

	bonus->sprite = sprite_create("res/bonus.txt");

	if(strcmp(bonus_type, "GravityFlip") == 0)
	{
		bonus->type = BONUSTYPE_GRAVITY_FLIP;
		sprite_set_animation(bonus->sprite, "GravityFlip");
	}
	else if(strcmp(bonus_type, "Walljump") == 0)
	{
		bonus->type = BONUSTYPE_WALLJUMP;
		sprite_set_animation(bonus->sprite, "Walljump");
	}
	else if(strcmp(bonus_type, "Climbing") == 0)
	{
		bonus->type = BONUSTYPE_CLIMBING;
		sprite_set_animation(bonus->sprite, "Climbing");
	}
	else if(strcmp(bonus_type, "Jetpack") == 0)
	{
		bonus->type = BONUSTYPE_JETTING;
		sprite_set_animation(bonus->sprite, "Jetpack");
	}
	else if(strcmp(bonus_type, "WaterLandFlip") == 0)
	{
		bonus->type = BONUSTYPE_WATER_LAND_FLIP;
		sprite_set_animation(bonus->sprite, "WaterLandFlip");
	}
	else
	{
		fprintf(stderr, "Trying to set unknown bonus type: '%s'\n", bonus_type);
		exit(EXIT_FAILURE);
	}

	if(args->addfunc != NULL)
		args->addfunc(gameobject);

	return NULL;
}
