//##############################################################################
// siphon
//
// player.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <SDL.h>
#include <assert.h>
#include <math.h>

#include "player.h"
#include "bonus.h"
#include "config.h"
#include "main.h"
#include "game.h"
#include "sound.h"
#include "vector.h"
#include "resources.h"
#include "viewport.h"
#include "transition.h"
#include "levelselect.h"

//##############################################################################
// GameObject registration
//##############################################################################
CONSTRUCTOR(player_const)
{
	registry_register("Player", &player_create);
}

//##############################################################################
// Static functions
//##############################################################################
static bool is_in_water(Map * map, int x, int y)
{
	int tilenum;
	Vector tl;
	tl.x = x >> map->tileshift;
	tl.y = y >> map->tileshift;
	tilenum = map_get_tilenum(map, tl);
	if(tilenum == 1 || tilenum == 2)
		return TRUE;
	return FALSE;
}

static void undo_rope_drop(void * gameobject, Event * event)
{
	GameObject * go = (GameObject*)gameobject;
	GameObject * rope = (GameObject*)event->p1;
	player_pickup_rope(go, rope);
}

static void player_destroy(GameObject * vp_player)
{
	Player * player = (Player*)vp_player;

	assert(vp_player && vp_player->type & PLAYER);

	sprite_destroy(player->sprite);

	siphon_free(player);
}

static void player_drop_rope(GameObject * go_player)
{
	Player * player = (Player*)go_player;
	Rope * rope = NULL;
	Uint8 * keystate = NULL;

	assert(go_player && go_player->type & PLAYER);
	assert(player->holding && player->holding->type & ROPE);

	rope = (Rope*)player->holding;

	keystate = game_get_keystate(go_player);

	if(keystate[SDLK_DOWN])
	{
		player->holding->req_vel.x = TOFIX(0);
		player->holding->req_vel.y = TOFIX(0);
	}
	else if(keystate[SDLK_UP] && player->direction == LEFT)
	{
		player->holding->req_vel.x -= TOFIX(1);
		player->holding->req_vel.y -= TOFIX(6);
	}
	else if(keystate[SDLK_UP])
	{
		player->holding->req_vel.x += TOFIX(1);
		player->holding->req_vel.y -= TOFIX(6);
	}
	else if(player->direction == LEFT)
	{
		player->holding->req_vel.x -= TOFIX(5) / 2;
		player->holding->req_vel.y -= TOFIX(4);
	}
	else
	{
		player->holding->req_vel.x += TOFIX(5) / 2;
		player->holding->req_vel.y -= TOFIX(4);
	}

	player->holding->size.x = 4;
	player->holding->size.y = 4;

	rope->holder = NULL;
	player->holding = NULL;

	gameobject_push_event(go_player, rope, NULL, &undo_rope_drop);
}

static void undo_rope_pickup(void * gameobject, Event * event)
{
	GameObject * go = (GameObject*)gameobject;
	player_drop_rope(go);
}

static void player_actionable_helper(GameObject * vp_player, GameObject * other)
{
	assert(vp_player && vp_player->type & PLAYER);
	assert(other);

	if(other->type & BONUS)
	{
		Bonus * bonus = (Bonus*)other;
		switch(bonus->type)
		{
			case BONUSTYPE_GRAVITY_FLIP:
				//config.flip_map_horizontally = !config.flip_map_horizontally;
				config.flip_gravity = !config.flip_gravity;
				break;
			case BONUSTYPE_WALLJUMP:
				game_set_message("You can now do walljumps");
				config.player_can_walljump = TRUE;
				break;
			case BONUSTYPE_CLIMBING:
				game_set_message("You can now climb walls");
				config.player_can_climb = TRUE;
				break;
			case BONUSTYPE_JETTING:
				game_set_message("You can now use a jetpack");
				config.player_can_jet = TRUE;
				break;
			case BONUSTYPE_WATER_LAND_FLIP:
				config.flip_water_land = TRUE;
				break;
		}
	}
}

static void player_obstacle_helper(GameObject * vp_player, GameObject * other)
{
	Player * player = (Player*)vp_player;

	assert(other);
	assert(vp_player && vp_player->type & PLAYER);

	if(other->type & SPIKES)
	{
		player->is_dead = TRUE;
	}
	else if(other->type & FINISH)
	{
		player->has_finished = TRUE;
	}
}

static void player_pickup_helper(GameObject * vp_player, GameObject * other)
{
	Player * player = (Player*)vp_player;

	assert(other);
	assert(vp_player && vp_player->type & PLAYER);

	if(other->type & ROPE)
	{
		Rope * rope = (Rope*)other;
		if(player->holding == NULL && rope->holder == NULL)
		{
			player_pickup_rope(vp_player, other);
			player->trying_pickup = FALSE;
		}
	}
}

static void player_logic(GameObject * vp_player)
{
	Player * player = (Player*)vp_player;
	GameObject * go = vp_player;
	Uint8 * keystate = NULL;
	Vector tl;
	bool climbing_wall = FALSE;
	int touch_ground, touch_left, touch_right, touch_top, in_water;
	bool throw_just_pressed, jump_just_pressed, pickup_just_pressed;
	bool action_just_pressed;
	FixedVector gravity;
	unsigned int i = 0;

	go->req_vel = go->vel;
	go->req_pos = go->pos;

	assert(vp_player && vp_player->type & PLAYER);

	if(player->is_dead)
		keystate = game_get_keystate(NULL);
	else
		keystate = game_get_keystate(go);

	touch_ground = gameobject_collides(go, COLLISION_DOWN);
	touch_left  = gameobject_collides(go, COLLISION_LEFT);
	touch_right = gameobject_collides(go, COLLISION_RIGHT);
	touch_top = gameobject_collides(go, COLLISION_UP);

	if(config.flip_gravity)
	{
		bool temp = touch_ground;
		touch_ground = touch_top;
		touch_top = temp;
	}

	//fprintf(stdout, "%d, %d\n", TOINT(go->pos.x), TOINT(go->pos.y));

	tl.x = TOINT(go->pos.x) + go->size.x / 2;
	tl.y = TOINT(go->pos.y) + go->size.y - 2;
	tl.x >>= go->map->tileshift;
	tl.y >>= go->map->tileshift;
	in_water = map_get_tilenum(go->map, tl);
	in_water = (in_water == 1 || in_water == 2) ? TRUE : FALSE;

	if(config.flip_water_land)
		in_water = !in_water;

	player->aircontrol_lock = MAX(0, player->aircontrol_lock - 1);
	if(touch_ground || touch_left || touch_right || touch_top || in_water)
	{
		player->aircontrol_lock = 0;
	}

	throw_just_pressed = !player->previous_throw && keystate[SDLK_f];
	player->previous_throw = keystate[SDLK_f];

	jump_just_pressed = !player->previous_jump && keystate[SDLK_a];
	player->previous_jump = keystate[SDLK_a];

	pickup_just_pressed = !player->previous_pickup && keystate[SDLK_d];
	player->previous_pickup = keystate[SDLK_d];

	action_just_pressed = !player->previous_action && keystate[SDLK_DOWN];
	player->previous_action = keystate[SDLK_DOWN];


	if(pickup_just_pressed && player->holding == NULL)
	{
		player->trying_pickup = TRUE;
	}
	if(!keystate[SDLK_d])
		player->trying_pickup = FALSE;

	player->has_finished = FALSE;
	game_do_colliders(go, &player_obstacle_helper);


	if(action_just_pressed)
	{
		game_do_colliders(go, &player_actionable_helper);
	}
	else if(player->trying_pickup && player->holding == NULL)
	{
		game_do_colliders(go, &player_pickup_helper);
	}
	else if(throw_just_pressed && player->holding != NULL)
	{
		player_drop_rope(vp_player);
	}

	if(!player->jetting && jump_just_pressed &&
		!touch_ground && config.player_can_jet)
	{
		player->jetting = TRUE;
	}

	if(!keystate[SDLK_a] || in_water)
	{
		player->jetting = FALSE;
	}

	gravity.x = TOFIX(0) / 2;
	gravity.y = TOFIX(1) / 2;

	if(in_water)
		gravity.y = TOFIX(1) / 40;

	if(in_water)
	{
		player->ticks++;
		if(player->ticks > 60)
		{
			if(player->bubble_count < sizeof(player->bubbles) / sizeof(Vector))
			{
				player->bubbles[player->bubble_count].x = TOINT(go->pos.x);
					if(player->direction == RIGHT)
						player->bubbles[player->bubble_count].x += 6;
				player->bubbles[player->bubble_count].y = TOINT(go->pos.y);
				player->bubble_count++;
			}
			player->ticks = 0;
		}

		if(go->req_vel.x > 0)
		{
			sprite_set_animation(player->sprite, "SwimRight");
			player->direction = RIGHT;
		}
		else if(go->req_vel.x < 0)
		{
			sprite_set_animation(player->sprite, "SwimLeft");
			player->direction = LEFT;
		}
		else if(player->direction == RIGHT)
		{
			if(go->req_vel.y < 0)
				sprite_set_animation(player->sprite, "SwimRight");
			else
				sprite_set_animation(player->sprite, "StrokeRight");
		}
		else
		{
			if(go->req_vel.y < 0)
				sprite_set_animation(player->sprite, "SwimLeft");
			else
				sprite_set_animation(player->sprite, "StrokeLeft");
		}
	}
	else if(touch_ground && !player->jetting)
	{
		if(go->req_vel.x > 0)
		{
			sprite_set_animation(player->sprite, "WalkRight");
			player->direction = RIGHT;
		}
		else if(go->req_vel.x < 0)
		{
			sprite_set_animation(player->sprite, "WalkLeft");
			player->direction = LEFT;
		}
		else if(player->direction == RIGHT)
			sprite_set_animation(player->sprite, "StandRight");
		else
			sprite_set_animation(player->sprite, "StandLeft");
	}
	else if(!in_water && keystate[SDLK_a] && player->jetting)
	{
		if(config.flip_gravity)
		{
			if(go->req_vel.y < TOFIX(2))
				go->req_vel.y += TOFIX(2) / 3;
		}
		else
		{
			if(go->req_vel.y > TOFIX(-2))
				go->req_vel.y -= TOFIX(2) / 3;
		}
		if(go->req_vel.x > 0)
		{
			sprite_set_animation(player->sprite, "JumpRight");
			player->direction = RIGHT;
		}
		else if(go->req_vel.x < 0)
		{
			sprite_set_animation(player->sprite, "JumpLeft");
			player->direction = LEFT;
		}
		else if(player->direction == RIGHT)
			sprite_set_animation(player->sprite, "JumpRight");
		else
			sprite_set_animation(player->sprite, "JumpLeft");
	}
	else if(go->req_vel.y <= 0)
	{
		if(go->req_vel.x > 0)
		{
			sprite_set_animation(player->sprite, "JumpRight");
			player->direction = RIGHT;
		}
		else if(go->req_vel.x < 0)
		{
			sprite_set_animation(player->sprite, "JumpLeft");
			player->direction = LEFT;
		}
		else if(player->direction == RIGHT)
			sprite_set_animation(player->sprite, "JumpRight");
		else
			sprite_set_animation(player->sprite, "JumpLeft");
	}
	else if(go->req_vel.y > 0)
	{
		if(go->req_vel.x > 0)
		{
			sprite_set_animation(player->sprite, "FallRight");
			player->direction = RIGHT;
		}
		else if(go->req_vel.x < 0)
		{
			sprite_set_animation(player->sprite, "FallLeft");
			player->direction = LEFT;
		}
		else if(player->direction == RIGHT)
			sprite_set_animation(player->sprite, "FallRight");
		else
			sprite_set_animation(player->sprite, "FallLeft");
	}

	if(!in_water && !player->jetting)
	{
		if(touch_right && keystate[SDLK_RIGHT] && config.player_can_climb)
		{
			player->direction = RIGHT;
			if(keystate[SDLK_UP])
			{
				if(config.flip_gravity)
					go->req_vel.y = TOFIX(3);
				else
					go->req_vel.y = TOFIX(-3);
				sprite_set_animation(player->sprite, "ClimbRight");
				climbing_wall = TRUE;
			}
			else
			{
				if(config.flip_gravity)
					go->req_vel.y = MAX(TOFIX(1) / -2, go->req_vel.y);
				else
					go->req_vel.y = MIN(TOFIX(1) / 2, go->req_vel.y);
				sprite_set_animation(player->sprite, "SlideRight");
			}
		}
		else if(touch_left && keystate[SDLK_LEFT] && config.player_can_climb)
		{
			player->direction = LEFT;
			if(keystate[SDLK_UP])
			{
				if(config.flip_gravity)
					go->req_vel.y = TOFIX(3);
				else
					go->req_vel.y = TOFIX(-3);
				sprite_set_animation(player->sprite, "ClimbLeft");
				climbing_wall = TRUE;
			}
			else
			{
				if(config.flip_gravity)
					go->req_vel.y = MAX(TOFIX(1) / -2, go->req_vel.y);
				else
					go->req_vel.y = MIN(TOFIX(1) / 2, go->req_vel.y);
				sprite_set_animation(player->sprite, "SlideLeft");
			}
		}
	}

	if(jump_just_pressed && config.flip_gravity)
	{
		if(in_water)
			go->req_vel.y += TOFIX(1);
		else if(touch_ground)
		{
			go->req_vel.y = TOFIX(6);
			sound_play_sample(SOUND_JUMP);
		}
		else if(touch_right && config.player_can_walljump)
		{
			player->aircontrol_lock = 15;
			go->req_vel.y = TOFIX(6);
			go->req_vel.x = TOFIX(-7);
		}
		else if(touch_left && config.player_can_walljump)
		{
			player->aircontrol_lock = 15;
			go->req_vel.y = TOFIX(6);
			go->req_vel.x = TOFIX(7);
		}
	}
	else if(jump_just_pressed)
	{
		if(in_water)
			go->req_vel.y += TOFIX(-1);
		else if(touch_ground)
		{
			go->req_vel.y = TOFIX(-6);
			sound_play_sample(SOUND_JUMP);			
		}
		else if(touch_right && config.player_can_walljump)
		{
			player->aircontrol_lock = 15;
			go->req_vel.y = TOFIX(-6);
			go->req_vel.x = TOFIX(-7);
		}
		else if(touch_left && config.player_can_walljump)
		{
			player->aircontrol_lock = 15;
			go->req_vel.y = TOFIX(-6);
			go->req_vel.x = TOFIX(7);
		}
	}

	if(config.flip_gravity)
	{
		gravity.x = -gravity.x;
		gravity.y = -gravity.y;
	}

	go->req_vel.x += gravity.x;
	go->req_vel.y += gravity.y;

	if(((!config.flip_left_right && keystate[SDLK_LEFT]) ||
		(config.flip_left_right && keystate[SDLK_RIGHT])) &&
			player->aircontrol_lock == 0)
	{
		if(keystate[SDLK_s] && player->holding != NULL &&
			player->holding->type & ROPE)
		{
			go->req_vel.x += TOFIX(-1) / 12;
		}
		else
		{
			go->req_vel.x += TOFIX(-1) / 3 * 2;
		}
	}

	if(((!config.flip_left_right && keystate[SDLK_RIGHT]) ||
		(config.flip_left_right && keystate[SDLK_LEFT])) &&
			player->aircontrol_lock == 0)
	{
		if(keystate[SDLK_s] && player->holding != NULL &&
			player->holding->type & ROPE)
		{
			go->req_vel.x += TOFIX(1) / 12;
		}
		else
		{
			go->req_vel.x += TOFIX(1) / 3 * 2;
		}
	}

	// Retract rope (Move player towards neirest rope node)
	if(player->holding != NULL && player->holding->type & ROPE)
	{
		Rope * rope = (Rope*)player->holding;
		if(keystate[SDLK_s] && !climbing_wall)
		{
			float xdiff;
			float ydiff;
			float length;
			CableNode * node = rope->start.next;

			if(node != NULL)
			{
				xdiff = node->pos.x - TOINT(go->pos.x) - 4;
				ydiff = node->pos.y - TOINT(go->pos.y) - 4;
			}
			else
			{
				xdiff = TOINT(rope->endpoint->pos.x) - TOINT(go->pos.x) - 4;
				ydiff = TOINT(rope->endpoint->pos.y) - TOINT(go->pos.y) - 4;
			}

			length = sqrt(xdiff * xdiff + ydiff * ydiff);

			if(length > 0.0f)
			{
				float xmod = (xdiff / length);
				float ymod = (ydiff / length);

				int f_xmod = (int)(xmod * 4096);
				int f_ymod = (int)(ymod * 4096);


				if(config.flip_rope)
				{
					go->req_vel.x -= f_xmod;
					go->req_vel.y -= f_ymod;
				}
				else
				{
					go->req_vel.x += f_xmod;
					go->req_vel.y += f_ymod;
				}
			}
		}
	}

	if(config.flip_gravity)
	{
		if(touch_ground)
			go->req_vel.y = MAX(go->req_vel.y, 0);
		if(touch_top)
			go->req_vel.y = MIN(go->req_vel.y, 0);
	}
	else
	{
		if(touch_ground)
			go->req_vel.y = MIN(go->req_vel.y, 0);
		if(touch_top)
			go->req_vel.y = MAX(go->req_vel.y, 0);
	}
	if(touch_left)
		go->req_vel.x = MAX(go->req_vel.x, 0);
	if(touch_right)
		go->req_vel.x = MIN(go->req_vel.x, 0);

	// Terminal velocities
	if(in_water)
	{
		if(go->req_vel.x > 0)
			go->req_vel.x -= MIN(go->req_vel.x, TOFIX(1) / 24);
		else if(go->req_vel.x < 0)
			go->req_vel.x -= MAX(go->req_vel.x, TOFIX(-1) / 24);

		go->req_vel.y = MIN(TOFIX(1), go->req_vel.y);
		go->req_vel.y = MAX(TOFIX(-2), go->req_vel.y);
		go->req_vel.x = MIN(TOFIX(1), go->req_vel.x);
		go->req_vel.x = MAX(TOFIX(-1), go->req_vel.x);
	}
	else
	{
		// Horizontal Friction (only if touching the ground)
		if((touch_ground) && go->req_vel.x > 0)
			go->req_vel.x -= MIN(go->req_vel.x, TOFIX(1) / 4);
		else if((touch_ground) && go->req_vel.x < 0)
			go->req_vel.x -= MAX(go->req_vel.x, TOFIX(-1) / 4);

		go->req_vel.y = MIN(TOFIX(5), go->req_vel.y);
		go->req_vel.y = MAX(TOFIX(-8), go->req_vel.y);
	}

	// Don't limit horizontal maximum speed as much when retracting rope
	if(keystate[SDLK_s] && player->holding != NULL &&
		player->holding->type & ROPE)
	{
		go->req_vel.x = MIN(TOFIX(4), go->req_vel.x);
		go->req_vel.x = MAX(TOFIX(-4), go->req_vel.x);
	}
	else
	{
		go->req_vel.x = MIN(TOFIX(2), go->req_vel.x);
		go->req_vel.x = MAX(TOFIX(-2), go->req_vel.x);
	}

	go->req_pos.x += go->req_vel.x;
	go->req_pos.y += go->req_vel.y;

	if(SDL_GetKeyState(NULL)[SDLK_r])
	{
		gameobject_pop_history(go);
		gameobject_pop_event(go);
	}

	if(player->is_dead)
	{
		sprite_set_animation(player->sprite, "Dead");
	}

	if(player->jetting)
	{
		int i = 0;
		for(i = 0; i < 10; i++)
		{
			unsigned int count = player->particle_count;
			Particle * particles = player->particles;
			if(count >= sizeof(player->particles) / sizeof(Particle))
				break;
			if(player->direction == RIGHT)
			{
				particles[count].x = go->pos.x + TOFIX(3);
				particles[count].y = go->pos.y + TOFIX(4);
			}
			else
			{
				particles[count].x = go->pos.x + TOFIX(go->size.x) - TOFIX(3);
				particles[count].y = go->pos.y + TOFIX(4);
			}
			particles[count].lifetime = 15;
			particles[count].xvel = (5 - (rand() % 10)) * (TOFIX(2) / 20);
			particles[count].yvel = TOFIX(1) / 2 + ((rand() % 20)) *
				(TOFIX(2) / 20);
			player->particle_count++;
		}
	}

	for(i = 0; i < player->particle_count; i++)
	{
		Particle * particles = player->particles;
		if(config.flip_gravity)
		{
			particles[i].x -= particles[i].xvel;
			particles[i].y -= particles[i].yvel;
		}
		else
		{
			particles[i].x += particles[i].xvel;
			particles[i].y += particles[i].yvel;
		}
		particles[i].lifetime--;
		if(particles[i].lifetime <= 0)
		{
			particles[i].x = particles[player->particle_count-1].x;
			particles[i].y = particles[player->particle_count-1].y;
			particles[i].lifetime =
				particles[player->particle_count-1].lifetime;
			i--;
			player->particle_count--;
		}
	}

	for(i = 0; i < player->bubble_count; i++)
	{
		if(config.flip_gravity)
			player->bubbles[i].y += 1;
		else
			player->bubbles[i].y -= 1;

		if(!is_in_water(go->map, player->bubbles[i].x, player->bubbles[i].y))
		{
			player->bubble_count--;
			player->bubbles[i] = player->bubbles[player->bubble_count];
		}
	}
}

static void player_draw(GameObject * vp_player, int offset_x, int offset_y)
{
	Player * player = (Player*)vp_player;
	GameObject * go = vp_player;
	unsigned int i = 0;
	//int a;
	int b;

	assert(vp_player && vp_player->type & PLAYER);

	//a = (TOINT(go->pos.x) - offset_x) * 2 + go->size.x;
	b = (TOINT(go->pos.y) - offset_y) * 2 + go->size.y;

	glColor3f(1.0f, 1.0f, 1.0f);

	glBegin(GL_POINTS);
	for(i = 0; i < player->bubble_count; i++)
	{
		Vector * bubble = player->bubbles + i;
		glVertex2f(bubble->x - offset_x, bubble->y - offset_y);
	}
	glEnd();

	glBindTexture(GL_TEXTURE_2D, *player->smoke_particle);
	glBegin(GL_QUADS);
	for(i = 0; i < player->particle_count; i++)
	{
		Particle * particle = player->particles + i;
		float s;
		glColor4f(1.0f, 1.0f, 1.0f, (0.5f * (particle->lifetime / 15.0f)) );

		s = 3.0f * (2.0f - (particle->lifetime / 15.0f));

		glTexCoord2f(0.0f, 0.0f);
		glVertex2d(TOINT(particle->x)-offset_x-s,TOINT(particle->y)-offset_y-s);
		glTexCoord2f(1.0f, 0.0f);
		glVertex2d(TOINT(particle->x)-offset_x+s,TOINT(particle->y)-offset_y-s);
		glTexCoord2f(1.0f, 1.0f);
		glVertex2d(TOINT(particle->x)-offset_x+s,TOINT(particle->y)-offset_y+s);
		glTexCoord2f(0.0f, 1.0f);
		glVertex2d(TOINT(particle->x)-offset_x-s,TOINT(particle->y)-offset_y+s);
	}
	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	if(config.flip_gravity)
	{
		glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
		glTranslatef(0.0f, -b, 0.0f);
	}

	sprite_draw(player->sprite, TOINT(go->pos.x) - offset_x,
		TOINT(go->pos.y) - offset_y);

	if(config.flip_gravity)
	{
		glTranslatef(0.0f, b, 0.0f);
		glRotatef(-180.0f, 1.0f, 0.0f, 0.0f);
	}
}

//##############################################################################
// Exported functions
//##############################################################################
GameObject * player_create(const CreateArgs * args)
{
	Player * player = (Player*)siphon_malloc(sizeof(Player));
	GameObject * gameobject = (GameObject*)player;

	Vector position = properties_get_vector(args->properties, "position");

	memset(player, 0, sizeof(Player));
	gameobject->pos = vector_to_fixed(position);
	gameobject->req_pos = vector_to_fixed(position);
	gameobject->map = args->map;
	gameobject->type = PLAYER;
	gameobject->size.x = 7;
	gameobject->size.y = 8;
	gameobject->logic = &player_logic;
	gameobject->draw = &player_draw;
	gameobject->destroy = &player_destroy;

	player->is_dead = FALSE;
	player->has_finished = FALSE;
	player->ticks = 0;
	player->previous_action = FALSE;
	player->previous_throw = FALSE;
	player->previous_jump = FALSE;
	player->previous_pickup = FALSE;
	player->trying_pickup = FALSE;
	player->aircontrol_lock = 0;
	player->jetting = FALSE;
	player->direction = RIGHT;

	player->particle_count = 0;
	player->bubble_count = 0;

	player->smoke_particle = resources_get_gl_image("res/smoke.png");

	player->sprite = sprite_create("res/player.txt");
	player->holding = NULL;
	sprite_set_animation(player->sprite, "StandRight");

	if(args->addfunc != NULL)
		args->addfunc(gameobject);

	return NULL;
}

void player_pickup_rope(GameObject * go_player, GameObject * go_pickup)
{
	Player * player = (Player*)go_player;

	assert(go_player && go_player->type & PLAYER);
	assert(go_pickup);
	assert(player->holding == NULL);

	if(go_pickup->type & ROPE)
	{
		Rope * rope = (Rope*)go_pickup;
		assert(rope->holder == NULL);
		rope->holder = go_player;
	}

	go_pickup->size.x = 1;
	go_pickup->size.y = 1;

	player->holding = go_pickup;

	gameobject_push_event(go_player, NULL, NULL, &undo_rope_pickup);
}
