//##############################################################################
// siphon
//
// rope.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "game.h"
#include "rope.h"
#include "bresenham.h"
#include "opengl_helper.h"
#include "resources.h"
#include "sound.h"
#include "bresenham.h"
#include "font.h"
#include "vector.h"

//##############################################################################
// Function prototypes
//##############################################################################
static GameObject * rope_create_helper(const CreateArgs * args);

//##############################################################################
// GameObject registration
//##############################################################################
CONSTRUCTOR(rope_const)
{
	registry_register("Rope", &rope_create_helper);
}

//##############################################################################
// Static variables
//##############################################################################
static GLuint * m_texture = NULL;

//##############################################################################
// Static functions
//##############################################################################
static bool tilecol(Map * map, int x, int y)
{
	Vector tl;
	int tilenum;
	tl.x = x;
	tl.y = y;
	tilenum = map_get_tilenum(map, tl);
	if(tilenum == -1 || (tilenum >= 4))
		return TRUE;
	return FALSE;
}

static BresenhamAction cable_collides_helper(FIXED_POINT x, FIXED_POINT y, void * data)
{
	Vector tl;
	Map * map = (Map*)data;

	tl.x = TOINT(x) >> map->tileshift;
	tl.y = TOINT(y) >> map->tileshift;

	if(tilecol(map, tl.x, tl.y))
		return STOP;
	return NEXT;
}

static BresenhamAction cable_helper(FIXED_POINT x, FIXED_POINT y, void * data)
{
	Vector tl;
	CableNode * end = (CableNode*)data;

	tl.x = (TOINT(x) >> end->map->tileshift);
	tl.y = (TOINT(y) >> end->map->tileshift);

	if(tilecol(end->map, tl.x, tl.y))
	{
		CableNode * pip = NULL;

		int newx = 0;
		int newy = 0;

		int left = tilecol(end->map, tl.x-1, tl.y);
		int right = tilecol(end->map, tl.x+1, tl.y);
		int up = tilecol(end->map, tl.x, tl.y-1);
		int down = tilecol(end->map, tl.x, tl.y+1);
		int upleft = tilecol(end->map, tl.x-1,tl.y-1);
		int upright = tilecol(end->map, tl.x+1,tl.y-1);
		int downleft = tilecol(end->map, tl.x-1,tl.y+1);
		int downright = tilecol(end->map, tl.x+1,tl.y+1);


		if(TOINT(y) % end->map->tilesize < 8 &&
			TOINT(x) % end->map->tilesize < 8) // TopLeft
		{
			newx = (tl.x << end->map->tileshift) - 1;
			newy = (tl.y << end->map->tileshift) - 1;
			if(left || upleft || up)
				return NEXT;
		}
		else if(TOINT(y) % end->map->tilesize < 8 ) // TopRight
		{
			newx = ((tl.x+1) << end->map->tileshift) + 1;
			newy = (tl.y << end->map->tileshift) - 1;
			if(right || upright || up)
				return NEXT;
		}
		else if(TOINT(x) % end->map->tilesize < 8) // BottomLeft
		{
			newx = (tl.x << end->map->tileshift) - 1;
			newy = ((tl.y+1) << end->map->tileshift) + 1;
			if(left || downleft || down)
				return NEXT;
		}
		else // BottomRight
		{
			newx = ((tl.x+1) << end->map->tileshift) + 1;
			newy = ((tl.y+1) << end->map->tileshift) + 1;
			if(right || downright || down)
				return NEXT;
		}

		if(newx == end->pos.x && newy == end->pos.y)
			return NEXT;

		pip = (CableNode*)siphon_malloc(sizeof(CableNode));
		pip->pos.x = newx;
		pip->pos.y = newy;
		pip->map = end->map;
		pip->next = NULL;
		end->next = pip;
		return STOP;
	}
	return NEXT;
}

static BresenhamAction cable_box_helper(FIXED_POINT x, FIXED_POINT y, void * data)
{
	Vector tl;
	CableNode * end = (CableNode*)data;;

	tl.x = (TOINT(x) >> end->map->tileshift);
	tl.y = (TOINT(y) >> end->map->tileshift);

	if(tilecol(end->map, tl.x, tl.y))
	{
		CableNode * pip = NULL;

		int newx = -1;
		int newy = -1;

		int left = tilecol(end->map, tl.x-1, tl.y);
		int right = tilecol(end->map, tl.x+1, tl.y);
		int up = tilecol(end->map, tl.x, tl.y-1);
		int down = tilecol(end->map, tl.x, tl.y+1);
		int upleft = tilecol(end->map, tl.x-1,tl.y-1);
		int upright = tilecol(end->map, tl.x+1,tl.y-1);
		int downleft = tilecol(end->map, tl.x-1,tl.y+1);
		int downright = tilecol(end->map, tl.x+1,tl.y+1);

		if(TOINT(y) % end->map->tilesize < 8 &&
			TOINT(x) % end->map->tilesize < 8) // TopLeft
		{
			newx = (tl.x << end->map->tileshift) - 1;
			newy = (tl.y << end->map->tileshift) - 1;
			if(left || upleft || up)
				return NEXT;
		}
		else if(TOINT(y) % end->map->tilesize < 8 ) // TopRight
		{
			newx = ((tl.x+1) << end->map->tileshift) + 1;
			newy = (tl.y << end->map->tileshift) - 1;
			if(right || upright || up)
				return NEXT;
		}
		else if(TOINT(x) % end->map->tilesize < 8) // BottomLeft
		{
			newx = (tl.x << end->map->tileshift) - 1;
			newy = ((tl.y+1) << end->map->tileshift) + 1;
			if(left || downleft || down)
				return NEXT;
		}
		else // BottomRight
		{
			newx = ((tl.x+1) << end->map->tileshift) + 1;
			newy = ((tl.y+1) << end->map->tileshift) + 1;
			if(right || downright || down)
				return NEXT;
		}

		if(newx == end->pos.x && newy == end->pos.y)
			return NEXT;

		pip = (CableNode*)siphon_malloc(sizeof(CableNode));
		pip->pos.x = end->pos.x;
		pip->pos.y = end->pos.y;
		pip->map = end->map;
		pip->next = end->next;

		end->pos.x = newx;
		end->pos.y = newy;
		end->next = pip;

		return STOP;
	}
	return NEXT;
}

static void rope_draw(GameObject * vp_rope, int map_x, int map_y)
{
	Rope * rope = (Rope*)vp_rope;
	GameObject * gameobject = vp_rope;
	CableNode * node = NULL;

	assert(vp_rope && vp_rope->type & ROPE);

	if(rope->holder == NULL)
	{
		sprite_draw(rope->sprite, TOINT(gameobject->pos.x) -
			map_x, TOINT(gameobject->pos.y) - map_y);
	}

	if(!rope_has_nodes(gameobject) && (rope->holder != NULL &&
		rope->holder == rope->endpoint))
		return;

	if(rope->endpoint == NULL)
		return;

	rope->start.pos.x = TOINT(gameobject->pos.x) + (gameobject->size.x / 2);
	rope->start.pos.y = TOINT(gameobject->pos.y) + (gameobject->size.y / 2);
	rope->start.map = gameobject->map;

	node = &rope->start;

	if(m_texture == NULL)
		m_texture = resources_get_gl_image("res/rope.png");

	glBindTexture(GL_TEXTURE_2D, *m_texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glBegin(GL_QUADS);
	while(node)
	{
		float x, y;
		float length;
		float xend;
		float yend;

		if(node->next == NULL)
		{
			xend = TOINT(rope->endpoint->pos.x) + rope->endpoint->size.x / 2;
			yend = TOINT(rope->endpoint->pos.y) + rope->endpoint->size.y / 2;
		}
		else
		{
			xend = node->next->pos.x;
			yend = node->next->pos.y;
		}

		x = -(node->pos.y - yend);
		y = node->pos.x - xend;

		length = sqrt(x*x + y*y);

		if(length != 0.0f)
		{
			x /= length;
			y /= length;
		}

		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(node->pos.x - map_x - x, node->pos.y - map_y - y);
		glTexCoord2f(1.0f, 0.0f);
		glVertex2f(node->pos.x - map_x + x, node->pos.y - map_y + y);
		glTexCoord2f(1.0f, 1.0f);
		glVertex2f(xend - map_x + x, yend - map_y + y);
		glTexCoord2f(0.0f, 1.0f);
		glVertex2f(xend - map_x - x, yend - map_y - y);

		node = node->next;
	}
	glEnd();
	glBindTexture(GL_TEXTURE_2D, 0);
}

static void rope_destroy(GameObject * vp_rope)
{
	CableNode * node = NULL;
	Rope * rope = (Rope*)vp_rope;

	assert(vp_rope && vp_rope->type & ROPE);

	sprite_destroy(rope->sprite);

	node = rope->start.next;
	while(node)
	{
		CableNode * destroy = node;
		node = node->next;
		siphon_free(destroy);
	}
	siphon_free(vp_rope);
}

static void rope_physics(Rope * rope)
{
	GameObject * go = (GameObject*)rope;
	bool just_collided;
	FixedVector gravity;
	int touch_ground, touch_left, touch_right, touch_top;
	int current_collide;

	assert(rope && go->type & ROPE);

	touch_ground = gameobject_collides(go, COLLISION_DOWN);
	touch_left  = gameobject_collides(go, COLLISION_LEFT);
	touch_right = gameobject_collides(go, COLLISION_RIGHT);
	touch_top = gameobject_collides(go, COLLISION_UP);

	current_collide = (touch_ground && go->req_vel.y > 0) +
						(touch_left && go->req_vel.x < 0) +
						(touch_right && go->req_vel.x > 0) +
						(touch_top && go->req_vel.y < 0);

	just_collided = (current_collide > rope->previous_collide) &&
		rope->holder == NULL;
	rope->previous_collide = current_collide;

	if(just_collided && game_is_on_screen(go))
	{
		sound_play_sample(SOUND_THUMP);
	}

	go->req_pos = go->pos;

	gravity.x = TOFIX(0) / 4;
	gravity.y = TOFIX(1) / 4;

	if(touch_ground)
		go->req_vel.y = MIN(go->req_vel.y, 0);
	if(touch_top)
		go->req_vel.y = MAX(go->req_vel.y, 0);
	if(touch_left)
		go->req_vel.x = MAX(go->req_vel.x, 0);
	if(touch_right)
		go->req_vel.x = MIN(go->req_vel.x, 0);

	go->req_vel.x += gravity.x;
	go->req_vel.y += gravity.y;

	if((touch_ground) && go->req_vel.x > 0)
		go->req_vel.x -= MIN(go->req_vel.x, TOFIX(1) / 3);
	else if((touch_ground) && go->req_vel.x < 0)
		go->req_vel.x -= MAX(go->req_vel.x, TOFIX(-1) / 3);

	go->req_vel.x = MAX(TOFIX(-4), go->req_vel.x);
	go->req_vel.x = MIN(TOFIX(4), go->req_vel.x);

	go->req_vel.y = MAX(TOFIX(-7), go->req_vel.y);
	go->req_vel.y = MIN(TOFIX(7), go->req_vel.y);

	go->req_pos.x += go->req_vel.x;
	go->req_pos.y += go->req_vel.y;
}

static void rope_logic(GameObject * vp_rope)
{
	Rope * rope = (Rope*)vp_rope;
	GameObject * gameobject = vp_rope;

	assert(vp_rope && vp_rope->type & ROPE);
	assert(gameobject->logic_updated == FALSE);

	if(rope->holder != NULL && !rope->holder->logic_updated)
	{
		GameObject * holder = NULL;
		gameobject->logic_updated = TRUE;
		holder = rope->holder;
		holder->logic(holder);
		// The above call to a holder's logic could reset rope->holder,
		// so make a copy first, so we can set the logic_updated field
		holder->logic_updated = TRUE;
	}

	rope->start.pos.x = TOINT(gameobject->pos.x) + (gameobject->size.x / 2);
	rope->start.pos.y = TOINT(gameobject->pos.y) + (gameobject->size.y / 2);
	rope->start.map = gameobject->map;

	if(rope->endpoint != NULL)
	{
		CableNode * joint = NULL;

		// Even the cable movement will be multisampled
		unsigned int i = 0;
		for(i = 0; i < gameobject->movement_count; i++)
		{
			FixedVector pos = gameobject->movement[i];
			CableNode * joint = &rope->start;
			if(joint->next == NULL)
			{
				bresenham_line_do(
					rope->endpoint->pos.x + TOFIX(rope->endpoint->size.x / 2),
					rope->endpoint->pos.y +	TOFIX(rope->endpoint->size.y / 2),
					pos.x + TOFIX(2),
					pos.y + TOFIX(2),
					cable_helper, (void*)joint);
			}
			else if(joint->next != NULL)
			{
				bresenham_line_do(
					TOFIX(joint->next->pos.x),
					TOFIX(joint->next->pos.y),
					pos.x + TOFIX(gameobject->size.x / 2),
					pos.y + TOFIX(gameobject->size.y / 2),
					cable_box_helper, (void*)joint->next);
			}
		}
		for(i = 0; i < rope->endpoint->movement_count; i++)
		{
			FixedVector pos = rope->endpoint->movement[i];
			CableNode * joint = &rope->start;
			while(joint)
			{
				if(joint->next == NULL)
				{
					bresenham_line_do(TOFIX(joint->pos.x), TOFIX(joint->pos.y),
						pos.x + TOFIX(rope->endpoint->size.x / 2),
						pos.y + TOFIX(rope->endpoint->size.y / 2),
						cable_helper, (void*)joint);
					break;
				}
				joint = joint->next;
			}
		}

		// Remove unnessecary joints
		joint = &rope->start;
		while(joint->next)
		{
			if(joint->next->next == NULL)
			{
				if(!bresenham_line_do(TOFIX(joint->pos.x), TOFIX(joint->pos.y),
					rope->endpoint->pos.x + TOFIX(rope->endpoint->size.x / 2),
					rope->endpoint->pos.y + TOFIX(rope->endpoint->size.y / 2),
					cable_collides_helper,
					(void*)gameobject->map))
				{
					int x1 = joint->pos.x;
					int y1 = joint->pos.y;


					int ax = joint->next->pos.x - x1;
					int ay = joint->next->pos.y - y1;


					int bx = TOINT(rope->endpoint->pos.x) +
						(rope->endpoint->size.x / 2) - x1;
					int by = TOINT(rope->endpoint->pos.y) +
						(rope->endpoint->size.y / 2) - y1;

					int blen = bx * bx + by * by;

					int dp = ax * bx + ay * by;

					float projx = dp / (float)blen * bx;
					float projy = dp / (float)blen * by;

					int diffx = ax - projx;
					int diffy = ay - projy;

					int length = sqrt(diffx * diffx + diffy * diffy);

					if(length >= 16)
					{
						break;
					}

					siphon_free(joint->next);
					joint->next = NULL;
				}
				break;
			}

			joint = joint->next;
		}

		joint = &rope->start;
		if(joint->next != NULL && joint->next->next != NULL)
		{
			if(!bresenham_line_do(TOFIX(joint->pos.x), TOFIX(joint->pos.y),
				TOFIX(joint->next->next->pos.x),
				TOFIX(joint->next->next->pos.y),
				cable_collides_helper, (void*)gameobject->map))
			{
				int x1 = joint->next->next->pos.x;
				int y1 = joint->next->next->pos.y;


				int ax = joint->next->pos.x - x1;
				int ay = joint->next->pos.y - y1;

				int bx = joint->pos.x - x1;
				int by = joint->pos.y - y1;

				int blen = bx * bx + by * by;

				int dp = ax * bx + ay * by;

				float projx = dp / (float)blen * bx;
				float projy = dp / (float)blen * by;

				int diffx = ax - projx;
				int diffy = ay - projy;

				int length = sqrt(diffx * diffx + diffy * diffy);

				if(length < 16)
				{
					CableNode * newnext = joint->next->next;
					siphon_free(joint->next);
					joint->next = newnext;
				}
			}
		}
	}

	if(SDL_GetKeyState(NULL)[SDLK_r])
	{
		gameobject_pop_history(gameobject);
	}
	else
	{
		// If the ropebox is being carried dont do physics just match
		// the position of the holder
		if(rope->holder != NULL)
		{
			gameobject->req_pos = rope->holder->req_pos;
			gameobject->req_pos.x += TOFIX(rope->holder->size.x / 2);
			gameobject->req_pos.y += TOFIX(rope->holder->size.y / 2);
		}
		else
		{
			rope_physics(rope);
		}
	}
}

static GameObject * rope_create_helper(const CreateArgs * args)
{
	GameObject * rope_a = rope_create(args);
	GameObject * rope_b = rope_create(args);

	Vector position = properties_get_vector(args->properties, "position");
	Vector endposition = properties_get_vector(args->properties, "endposition");

	rope_a->pos = vector_to_fixed(position);
	rope_a->req_pos = vector_to_fixed(position);
	rope_a->map = args->map;

	rope_b->pos = vector_to_fixed(endposition);
	rope_b->req_pos = vector_to_fixed(endposition);
	rope_b->map = args->map;

	rope_attach(rope_a, rope_b);
	rope_attach(rope_b, rope_a);

	if(args->addfunc != NULL)
	{
		args->addfunc(rope_a);
		args->addfunc(rope_b);
	}

	return NULL;
}

//##############################################################################
// Exported functions
//##############################################################################
GameObject * rope_create(const CreateArgs * args)
{
	Rope * rope = (Rope*)siphon_malloc(sizeof(Rope));
	GameObject * gameobject = (GameObject*)&rope->gameobject;
	memset(rope, 0x0, sizeof(Rope));

	gameobject->type = ROPE;
	gameobject->size.x = 4;
	gameobject->size.y = 4;
	gameobject->destroy = &rope_destroy;
	gameobject->logic = &rope_logic;
	gameobject->draw = &rope_draw;

	rope->sprite = sprite_create("res/rope.txt");
	rope->previous_collide = 0;
	sprite_set_animation(rope->sprite, "Rope");

	return gameobject;
}

bool rope_has_nodes(GameObject * go_rope)
{
	Rope * rope = (Rope*)go_rope;
	assert(go_rope && go_rope->type & ROPE);
	return rope->start.next != NULL;
}

void rope_attach(GameObject * go_rope, GameObject * attachee)
{
	Rope * rope = (Rope*)go_rope;
	assert(go_rope && go_rope->type & ROPE);
	rope->endpoint = attachee;
}
