//##############################################################################
// siphon
//
// signpost.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <assert.h>
#include <SDL.h>

#include "signpost.h"
#include "game.h"
#include "config.h"

//##############################################################################
// GameObject registration
//##############################################################################
CONSTRUCTOR(signpost_const)
{
	registry_register("Signpost", &signpost_create);
}

//##############################################################################
// Static functions
//##############################################################################
static void signpost_destroy(GameObject * vp_signpost)
{
	Signpost * signpost = (Signpost*)vp_signpost;
	assert(vp_signpost && vp_signpost->type & SIGNPOST);
	sprite_destroy(signpost->sprite);
	siphon_free(signpost->text);
	siphon_free(signpost);
}

static void signpost_logic(GameObject * gameobject)
{
	FixedVector gravity;
	GameObject * go = gameobject;
	int touch_ground = gameobject_collides(go, COLLISION_DOWN);
	int touch_left = gameobject_collides(go, COLLISION_LEFT);
	int touch_right = gameobject_collides(go, COLLISION_RIGHT);
	int touch_top = gameobject_collides(go, COLLISION_UP);

	gameobject->req_pos = gameobject->pos;
	gameobject->req_vel = gameobject->vel;

	gravity.x = TOFIX(0) / 4;
	gravity.y = TOFIX(1) / 4;

	if(config.flip_gravity)
	{
		gravity.y = -gravity.y;
		gravity.x = -gravity.x;
	}

	if(touch_ground)
		go->req_vel.y = MIN(go->req_vel.y, 0);
	if(touch_top)
		go->req_vel.y = MAX(go->req_vel.y, 0);
	if(touch_left)
		go->req_vel.x = MAX(go->req_vel.x, 0);
	if(touch_right)
		go->req_vel.x = MIN(go->req_vel.x, 0);

	gameobject->req_vel.x += gravity.x;
	gameobject->req_vel.y += gravity.y;

	gameobject->req_vel.y = MAX(TOFIX(-7), gameobject->req_vel.y);
	gameobject->req_vel.y = MIN(TOFIX(7), gameobject->req_vel.y);

	gameobject->req_pos.x += gameobject->req_vel.x;
	gameobject->req_pos.y += gameobject->req_vel.y;
}

static void signpost_draw(GameObject * vp_signpost, int offset_x, int offset_y)
{
	Signpost * signpost = (Signpost*)vp_signpost;
	int b;
	assert(vp_signpost && vp_signpost->type & SIGNPOST);

	b = (TOINT(vp_signpost->pos.y) - offset_y) * 2 + vp_signpost->size.y;

	if(config.flip_gravity)
	{
		glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
		glTranslatef(0.0f, -b, 0.0f);
	}

	sprite_draw(signpost->sprite, TOINT(vp_signpost->pos.x) - offset_x,
		TOINT(vp_signpost->pos.y) - offset_y);

	if(config.flip_gravity)
	{
		glTranslatef(0.0f, b, 0.0f);
		glRotatef(-180.0f, 1.0f, 0.0f, 0.0f);
	}
}

//##############################################################################
// Exported functions
//##############################################################################
GameObject * signpost_create(const CreateArgs * args)
{
	Signpost * signpost = (Signpost*)siphon_malloc(sizeof(Signpost));
	GameObject * gameobject = (GameObject*)signpost;

	Vector position = properties_get_vector(args->properties, "position");
	const char * text = properties_get_value(args->properties, "text");

	memset(signpost, 0, sizeof(Signpost));
	gameobject->map = args->map;
	gameobject->pos = vector_to_fixed(position);
	gameobject->req_pos = vector_to_fixed(position);
	gameobject->type = SIGNPOST;
	gameobject->size.x = 11;
	gameobject->size.y = 10;
	gameobject->logic = &signpost_logic;
	gameobject->draw = &signpost_draw;
	gameobject->destroy = &signpost_destroy;

	signpost->text = siphon_strdup(text);
	signpost->sprite = sprite_create("res/signpost.txt");
	sprite_set_animation(signpost->sprite, "Stand");

	if(args->addfunc != NULL)
		args->addfunc(gameobject);

	return NULL;
}

const char * signpost_get_message(GameObject * go_signpost)
{
	Signpost * signpost = (Signpost*)go_signpost;
	assert(go_signpost && go_signpost->type & SIGNPOST);
	return signpost->text;
}
