//##############################################################################
// siphon
//
// main.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <stdlib.h>
#include <SDL_mixer.h>
#include <assert.h>

#include "main.h"
#include "common.h"
#include "viewport.h"
#include "sound.h"
#include "config.h"
#include "registry.h"
#include "resources.h"
#include "frameratelimiter.h"

//##############################################################################
// Definitions
//##############################################################################
#define MAX_STATES 16

//##############################################################################
// Static variables
//##############################################################################
static bool m_running = TRUE;
static unsigned int m_gamestate = 0;
static unsigned int m_next_gamestate = 0;
static unsigned int m_states_count = 1;
static StateInterface m_states[MAX_STATES];

//##############################################################################
// Static functions
//##############################################################################
static void main_update_gamestate()
{
	if(m_gamestate == m_next_gamestate)
		return;

	if(m_states[m_gamestate].leave_state != NULL)
		m_states[m_gamestate].leave_state(m_states[m_next_gamestate].name);

	m_gamestate = m_next_gamestate;
	m_states[m_gamestate].enter_state();
}

static void main_handle_events()
{
	SDL_Event event;
	while(SDL_PollEvent(&event))
	{
		switch(event.type)
		{
			case SDL_QUIT:
				m_running = FALSE;
				break;
		}

		viewport_handle_event(&event);
		m_states[m_gamestate].handle_event(&event);
	}
}

static void main_initialize()
{
	// Compare linktime and compiletime versions of SDL
	SDL_version compile_version;
	const SDL_version * link_version = SDL_Linked_Version();
	SDL_VERSION(&compile_version);
	if(compile_version.major != link_version->major ||
		compile_version.minor != link_version->minor ||
		compile_version.patch != link_version->patch)
	{
		fprintf(stdout, "Compiled with SDL version: %d.%d.%d\n",
			compile_version.major,
			compile_version.minor,
			compile_version.patch);
		fprintf(stdout, "Running with SDL version: %d.%d.%d\n",
			link_version->major,
			link_version->minor,
			link_version->patch);
	}

	// Initialize SDL
	if(SDL_Init(SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO) == -1)
	{
		fprintf(stderr, "Failed to initialize SDL: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
}

static int main_get_state_index(const char * name)
{
	int i = 0;

	assert(name != NULL);
	
	for(i = 1; i < m_states_count; i++)
	{
		if(strcmp(m_states[i].name, name) == 0)
			return i;
	}
	
	fprintf(stderr, "Failed to find a state named '%s'\n", name);
	exit(EXIT_FAILURE);
	return -1;
}

//##############################################################################
// Exported functions
//##############################################################################
void main_register_state(StateInterface interface)
{
	// Prepare the m_states array (remember this gets executed before main)
	static bool first_invocation = TRUE;
	if(first_invocation)
	{
		memset(&m_states, 0x0, sizeof(m_states));
		first_invocation = FALSE;
	}

	// Make sure the interface is filled properly
	assert(interface.name != NULL);
	assert(interface.initialize != NULL);
	assert(interface.cleanup != NULL);
	assert(interface.logic != NULL);
	assert(interface.draw != NULL);
	assert(interface.handle_event != NULL);
	assert(interface.enter_state != NULL);
	assert(interface.leave_state != NULL);

	// Make sure that an already configured state does not get overwritten
	assert(m_states[m_states_count].initialize == NULL);

	// Is there enough space for another state?
	if(m_states_count == sizeof(m_states) / sizeof(m_states[0]))
	{
		fprintf(stderr, "Not enough space to register state '%s'\n",
			interface.name);
		exit(EXIT_FAILURE);
	}

	// Store the state
	m_states[(int)m_states_count] = interface;
	m_states_count++;
}

DrawFunc main_get_drawfunc(const char * name)
{
	int gamestate = main_get_state_index(name);
	assert(m_states[gamestate].draw != NULL);
	return m_states[gamestate].draw;
}

void main_set_running(bool running)
{
	m_running = running;
}

void main_set_gamestate(const char * name)
{
	// Make sure the gamestate gets set only once per frame
	assert(m_gamestate == m_next_gamestate);

	m_next_gamestate = main_get_state_index(name);
}

int main(int argc, char * argv[])
{
	int i = 0;

	main_initialize();
	config_initialize();
	viewport_initialize();
	resources_initialize();
	sound_initialize();

	// Initialize states
	for(i = 0; i < sizeof(m_states) / sizeof(m_states[0]); i++)
		if(m_states[i].initialize != NULL)
			m_states[i].initialize();

	// Set initial state
	if(argc == 2)
	{
		m_next_gamestate = main_get_state_index(argv[1]);
	}
	else
	{
		m_next_gamestate = main_get_state_index("intro");
	}

	while(m_running)
	{
		frameratelimiter_start();
		main_update_gamestate();
		main_handle_events();

		m_states[m_gamestate].logic();

		viewport_begin_draw();
		m_states[m_gamestate].draw();
		viewport_end_draw();

		frameratelimiter_stop();
		frameratelimiter_limit_rate();
	}

	// Cleanup states
	m_states[m_gamestate].leave_state(NULL);
	for(i = 0; i < sizeof(m_states) / sizeof(m_states[0]); i++)
		if(m_states[i].cleanup != NULL)
			m_states[i].cleanup();

	sound_cleanup();
	viewport_cleanup();
	registry_unregister();
	resources_unload();
	SDL_Quit();
	config_save();
	siphon_print_allocations();

	return EXIT_SUCCESS;
}
