//##############################################################################
// siphon
//
// map.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <SDL.h>

#include "map.h"
#include "vector.h"
#include "common.h"
#include "resources.h"
#include "viewport.h"
#include "opengl_helper.h"

#define VERTICES_PER_TILE 4
#define FLOATS_PER_TILE 8

//##############################################################################
// Type definitions
//##############################################################################
typedef struct
{
	float tlx, tly;
	float trx, try;
	float brx, bry;
	float blx, bly;
} FloatQuad;

//##############################################################################
// Static functions
//##############################################################################
///TODO: This should be revised
static void map_draw_background(Map * map, int map_x, int map_y)
{
	Vector size = viewport_get_size();
	int bottom = map->h * map->tilesize - viewport_get_size().y;
	float n = -0.15f + (1.0f / bottom) * map_y;
	int threshold = map->underground;

	n = MIN_F(0.5f, n);

	if(map_y + viewport_get_size().y > threshold)
	{
		glBegin(GL_QUADS);
			glColor4f(n+0.0f, n+0.15f, n+0.47f, 1.0f);
			glVertex2f(0, 0);
			glVertex2f(size.x, 0);
			glColor4f(n+0.39f, n+0.76f, n+1.0f, 1.0f);
			glVertex2f(size.x, threshold - map_y);
			glVertex2f(0, threshold - map_y);

			glVertex2f(0, threshold - map_y);
			glVertex2f(size.x, threshold - map_y);
			glColor4f(106.0f/255.0f, 78.0f / 255.0f, 42.0f / 255.0f, 1.0f);
			glVertex2f(size.x, threshold - map_y + 3);
			glVertex2f(0, threshold - map_y + 3);

			glVertex2f(0, threshold - map_y + 3);
			glVertex2f(size.x, threshold - map_y + 3);
			glColor4f(90.0f/255.0f, 62.0f / 255.0f, 26.0f / 255.0f, 1.0f);
			glVertex2f(size.x, size.y);
			glVertex2f(0, size.y);
		glEnd();
	}
	else
	{
		glBegin(GL_QUADS);
			glColor4f(n+0.0f, n+0.15f, n+0.47f, 1.0f);
			glVertex2f(0, 0);
			glColor4f(n+0.0f, n+0.15f, n+0.47f, 1.0f);
			glVertex2f(size.x, 0);
			glColor4f(n+0.39f, n+0.76f, n+1.0f, 1.0f);
			glVertex2f(size.x, size.y);
			glColor4f(n+0.39f, n+0.76f, n+1.0f, 1.0f);
			glVertex2f(0, size.y);
		glEnd();
	}

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
}

static void map_create_vbos(Map * map)
{
	GLsizei size = sizeof(GLfloat) * FLOATS_PER_TILE * map->w * map->h;
	map->vertices_vbo = resources_create_vbo(size, map->vertex_buffer);
	map->texcoords_vbo = resources_create_vbo(size, map->texcoord_buffer);
}

static void map_create_arrays(Map * map)
{
	int index = 0;
	int x, y;
	int x_tilecount = map->surface_size.x / map->tilesize;
	int y_tilecount = map->surface_size.y / map->tilesize;
	int shift = map->tileshift;
	float x_tilesize = 1.0f / x_tilecount;
	float y_tilesize = 1.0f / y_tilecount;
	FloatQuad * quadverts, * texcoords;

	map->vertex_buffer = (GLfloat*)siphon_malloc(sizeof(GLfloat) *
		FLOATS_PER_TILE * map->w * map->h);
	map->texcoord_buffer = (GLfloat*)siphon_malloc(sizeof(GLfloat) *
		FLOATS_PER_TILE * map->w * map->h);

	quadverts = (FloatQuad*)map->vertex_buffer;
	texcoords = (FloatQuad*)map->texcoord_buffer;

	for(y = 0; y < map->h; y++)
	{
		for(x = 0; x < map->w; x++)
		{
			int tile = map->tiles[x + y * map->w];
			int x_tile = tile % x_tilecount;
			int y_tile = tile / x_tilecount;

			quadverts[index].tlx = x << shift;
			quadverts[index].tly = y << shift;
			quadverts[index].trx = (x+1) << shift;
			quadverts[index].try = y << shift;
			quadverts[index].brx = (x+1) << shift;
			quadverts[index].bry = (y+1) << shift;
			quadverts[index].blx = x << shift;
			quadverts[index].bly = (y+1) << shift;

			texcoords[index].tlx = x_tile * x_tilesize;
			texcoords[index].tly = y_tile * y_tilesize;
			texcoords[index].trx = x_tile * x_tilesize + x_tilesize;
			texcoords[index].try = y_tile * y_tilesize;
			texcoords[index].brx = x_tile * x_tilesize + x_tilesize;
			texcoords[index].bry = y_tile * y_tilesize + y_tilesize;
			texcoords[index].blx = x_tile * x_tilesize;
			texcoords[index].bly = y_tile * y_tilesize + y_tilesize;

			index++;
		}
	}
}

static void parse_gameobject(Map * map, char * buffer)
{
	int i = 0;
	char * name = NULL;
	GameObjectPosition * position = NULL;
	Properties * properties = properties_create();

	// Read name of gameobject
	name = buffer;
	for(i = 0; buffer[i]; i++)
	{
		if(buffer[i] == ' ')
		{
			buffer[i] = '\0';
			buffer = &buffer[i+1];
			break;
		}
		else if(buffer[i+1] == '\0')
		{
			buffer = &buffer[i+2];
			break;
		}
	}

	// Was the name read?
	if(name == buffer)
	{
		fprintf(stderr, "Parsing error for line: %s\n", buffer);
		exit(EXIT_FAILURE);
	}

	// Skip over whitespace
	for(; buffer[0] == ' '; buffer++);

	// Read properties
	properties_add_from_string(properties, buffer);

	// Add gameobjectposition to map
	position = (GameObjectPosition*)siphon_malloc(sizeof(GameObjectPosition));
	position->properties = properties;
	position->name = siphon_strdup(name);
	position->next = map->gameobjects;
	map->gameobjects = position;
}

//##############################################################################
// Exported functions
//##############################################################################
Map * map_create(const char * name)
{
	Map * map = (Map*)siphon_malloc(sizeof(Map));
	FILE * file = NULL;
	SDL_Surface * surface = NULL;
	const char * tileset = NULL;
	char buffer[256];
	unsigned int x = 0;
	unsigned int y = 0;

	// Initialize map
	map->tickcount = 0;
	map->alive_players = 0;
	map->total_players = 0;
	map->finished_players = 0;
	map->mapname = siphon_strdup(name);
	map->gameobjects = NULL;
	map->properties = properties_create();

	// Try to open map file
	file = fopen(name, "r");
	if(file == NULL)
	{
		fprintf(stderr, "Failed to open map file: %s\n", name);
		exit(EXIT_FAILURE);
	}

	// Read map properties
	while(!feof(file))
	{
		char * result = fgets(buffer, sizeof(buffer) / sizeof(char), file);
		if(result == NULL)
		{
			fprintf(stderr, "%s: Failed to load map properties\n", name);
			exit(EXIT_FAILURE);
		}

		strip_whitespace(result);
		if(buffer[0] == '\n' || buffer[0] == '\r' || buffer[0] == '\0')
			continue;
		if(buffer[0] == '-')
			break;

		properties_add_from_string(map->properties, buffer);
	}

	// Initialize the tileset
	tileset = properties_get_value(map->properties, "tileset");
	map->tileset_texture = resources_get_gl_image(tileset);
	surface = resources_get_image(tileset);
	map->surface_size.x = surface->w;
	map->surface_size.y = surface->h;

	// Get required map properties
	map->underground = properties_get_int(map->properties, "groundlevel");
	map->tilesize = properties_get_int(map->properties, "tilesize");
	map->w = properties_get_int(map->properties, "mapwidth");
	map->h = properties_get_int(map->properties, "mapheight");

	// Read gameobjects, and positions
	while(!feof(file))
	{
		char * result = fgets(buffer, sizeof(buffer) / sizeof(char), file);
		if(result == NULL)
		{
			fprintf(stderr, "%s: Failed to load gameobjects\n", name);
			exit(EXIT_FAILURE);
		}

		strip_whitespace(result);
		if(buffer[0] == '\n' || buffer[0] == '\r' || buffer[0] == '\0')
			continue;
		if(buffer[0] == '-')
			break;

		parse_gameobject(map, buffer);
	}

	// Precalcalute shift amount to go from world coordinates to tile ones
	map->tileshift = -1;
	{
		int ts = map->tilesize;
		while(ts)
		{
			map->tileshift++;
			ts >>= 1;
		}
	}

	// Make sure tilesize is a power of two
	if((map->tilesize & (map->tilesize-1)) != 0)
	{
		fprintf(stderr, "%s: Tilesize is not a power of two\n", name);
		exit(EXIT_FAILURE);
	}

	// Allocate memory for map data, then read
	map->tiles = (unsigned int*)siphon_malloc(
		sizeof(unsigned int) * map->w * map->h);
	for(y = 0; y < map->h; y++)
	{
		for(x = 0; x < map->w; x++)
		{
			fscanf(file, "%u", &(map->tiles[y * map->w + x]));
		}
	}

	fclose(file);

	// Precalculate vertices
	map_create_arrays(map);
	map_create_vbos(map);

	return map;
}

void map_destroy(Map * map)
{
	assert(map != NULL);

	while(map->gameobjects)
	{
		GameObjectPosition * pos = map->gameobjects;
		map->gameobjects = pos->next;
		siphon_free(pos->name);
		properties_destroy(pos->properties);
		siphon_free(pos);
	}

	resources_destroy_vbo(map->vertices_vbo);
	resources_destroy_vbo(map->texcoords_vbo);
	properties_destroy(map->properties);
	siphon_free(map->texcoord_buffer);
	siphon_free(map->vertex_buffer);
	siphon_free(map->mapname);
	siphon_free(map->tiles);
	siphon_free(map);
}

int map_get_tilenum(Map * map, Vector tile_pos)
{
	int xtile = tile_pos.x;
	int ytile = tile_pos.y;

	if(xtile < 0 || xtile >= (int)map->w)
		return -1;
	if(ytile < 0 || ytile >= (int)map->h)
		return -1;

	return map->tiles[xtile + ytile * map->w];
}

void map_logic(Map * map)
{

}

void map_draw(Map * map, Vector offset)
{
	Vector size = viewport_get_size();

	int left = MAX(0, offset.x >> map->tileshift);
	int top = MAX(0, offset.y >> map->tileshift);
	int bottom = MIN(map->h - 1, (offset.y + size.y) >> map->tileshift);
	int count = MIN(map->w - left, (size.x >> map->tileshift) + 2);
	int index = left + top * map->w;

	map_draw_background(map, offset.x, offset.y);
	glTranslatef(-offset.x, -offset.y, 0.0f);
	glBindTexture(GL_TEXTURE_2D, *map->tileset_texture);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, *map->vertices_vbo);
	glVertexPointer(2, GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, *map->texcoords_vbo);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	for(; top <= bottom; top++)
	{
		glDrawArrays(GL_QUADS, index * VERTICES_PER_TILE,
			count * VERTICES_PER_TILE);
		index += map->w;
	}
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glBindTexture(GL_TEXTURE_2D, 0);
	glTranslatef(offset.x, offset.y, 0.0f);
}
