//##############################################################################
// siphon
//
// resources.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <SDL_image.h>
#include <assert.h>
#include <ctype.h>

#include "opengl_helper.h"
#include "resources.h"
#include "common.h"
#include "datastructures.h"

//##############################################################################
// Type definitions
//##############################################################################
typedef struct
{
	void * data;
	GLsizei size;
	GLuint handle;
} VboInfo;

//##############################################################################
// Static variables
//##############################################################################
static SDict * m_textures = NULL;
static SDict * m_surfaces = NULL;
static SDict * m_sprites = NULL;
static SList * m_vbos = NULL;

//##############################################################################
// Static functions
//##############################################################################
static SpriteInfo * resources_load_spriteinfo(const char * name)
{
	FILE * file = NULL;
	char buffer[256];
	GLint surface_width = 64, surface_height = 64;
	SpriteAnimation * current = NULL;
	SpriteInfo * spriteinfo = (SpriteInfo*)siphon_malloc(sizeof(SpriteInfo));
	spriteinfo->tex = NULL;
	spriteinfo->animation_count = 0;
	spriteinfo->animations = NULL;

	file = fopen(name, "r");

	// Make sure the file has been opened
	if(file == NULL)
	{
		fprintf(stderr, "Failed to open sprite file: %s.\n", name);
		exit(EXIT_FAILURE);
	}

	// Read texture filename
	if(fgets(buffer, 255, file) == NULL)
	{
		fprintf(stderr, "%s: Spritesheet filename is too long.\n", name);
		exit(EXIT_FAILURE);
	}
	strip_whitespace(buffer);
	spriteinfo->tex = resources_get_gl_image(buffer);

	// Retrieve surface width and height to use for precalculating texcoords
	glBindTexture(GL_TEXTURE_2D, *spriteinfo->tex);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH,
		&surface_width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT,
		&surface_height);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Read empty line
	if(fgets(buffer, 255, file) == NULL)
	{
		fprintf(stderr, "%s: Second line in sprite should be empty.\n", name);
		exit(EXIT_FAILURE);
	}
	strip_whitespace(buffer);
	if(buffer[0] != '\0')
	{
		fprintf(stderr, "%s: Second line in sprite should be empty.\n", name);
		exit(EXIT_FAILURE);
	}

	// Loop through animations
	while(!feof(file))
	{
		char * result = fgets(buffer, 255, file);
		strip_whitespace(buffer);

		if(result == NULL && feof(file))
		{
			break;
		}
		else if(result == NULL)
		{
			fprintf(stderr, "%s: Failed to fit line in buffer.\n", name);
			exit(EXIT_FAILURE);
		}

		if(isalpha(buffer[0]))
		{
			spriteinfo->animation_count++;
			spriteinfo->animations = (SpriteAnimation*)
				siphon_realloc(spriteinfo->animations,
					sizeof(SpriteAnimation) * spriteinfo->animation_count);
			current = &spriteinfo->animations[spriteinfo->animation_count-1];
			current->name = siphon_strdup(buffer);
			current->frame_count = 0;
			current->frames = NULL;
		}
		else if(buffer[0] == '\\')
		{
			sscanf(buffer, "\\ %d", &current->divisor);
		}
		else if(isdigit(buffer[0]))
		{
			int x, y, width, height;
			SpriteFrame * frame = NULL;
			current->frame_count++;
			current->frames = (SpriteFrame*)siphon_realloc(current->frames,
				sizeof(SpriteFrame) * current->frame_count);
			frame = &current->frames[current->frame_count-1];

			sscanf(buffer, "%d %d %d %d", &x, &y, &width, &height);

			frame->left = 1.0f / surface_width * x;
			frame->top = 1.0f / surface_height * y;
			frame->right = 1.0f / surface_width * (x+width);
			frame->bottom = 1.0f / surface_height * (y+height);
			frame->width = width;
			frame->height = height;
		}
	}

	fclose(file);
	return spriteinfo;
}

static SDL_Surface * resources_load_image(const char * name)
{
	SDL_Surface * surface = IMG_Load(name);
	if(surface == NULL)
	{
		fprintf(stderr, "%s: %s\n", name, IMG_GetError());

		// Create a dummy surface to indicate loading failed
		surface = SDL_CreateRGBSurface(SDL_SWSURFACE, 16, 16, 32,
			0x0, 0x0, 0x0, 0x0);

		if(surface == NULL)
		{
			fprintf(stderr, "Failed to create dummy texture for: '%s'\n", name);
			exit(EXIT_FAILURE);
		}

		assert(surface->format->BytesPerPixel == 4);
		if(surface->format->BytesPerPixel == 4)
		{
			int x = 0;
			unsigned int * pixels = (unsigned int*)surface->pixels;
			assert(surface->w == 16 && surface->h == 16);
			// Draw a red cross on the surface
			for(x = 0; x < 16; x++)
			{
				pixels[x + x * 16] = 0xFFFF0000;
				pixels[15-x + x * 16] = 0xFFFF0000;
			}
		}
	}

	if(surface != NULL)
	{
		SDL_Surface * swap = SDL_DisplayFormatAlpha(surface);
		if(swap != NULL)
		{
			SDL_FreeSurface(surface);
			surface = swap;
		}
	}
	return surface;
}

static void handle_opengl_error(const char * name, const char * description)
{
	char * error_string = NULL;
	GLenum error = glGetError();
	switch(error)
	{
		case(GL_NO_ERROR):
			return;
		case GL_INVALID_ENUM:
			error_string = "GL_INVALID_ENUM";
			break;
		case GL_INVALID_VALUE:
			error_string = "GL_INVALID_VALUE";
			break;
		case GL_INVALID_OPERATION:
			error_string = "GL_INVALID_OPERATION";
			break;
		case GL_STACK_OVERFLOW:
			error_string = "GL_STACK_OVERFLOW";
			break;
		case GL_STACK_UNDERFLOW:
			error_string = "GL_STACK_UNDERFLOW";
			break;
		case GL_OUT_OF_MEMORY:
			error_string = "GL_OUT_OF_MEMORY";
			break;
	}

	fprintf(stderr, "%s: OpenGL error: '%s' after %s\n", name,
		error_string, description);
	exit(EXIT_FAILURE);
}

static GLuint resources_load_gl_image(const char * name)
{
	int width, height;
	bool created_new = FALSE;
	SDL_Surface * surface = resources_get_image(name);
	GLint numberOfColors;
	GLenum textureFormat;
	GLuint texture;

	// Create image with width and height that are powers of 2
	width = surface->w-1;
	width = (width >> 1) | width;
	width = (width >> 2) | width;
	width = (width >> 4) | width;
	width = (width >> 8) | width;
	width = (width >> 16) | width;
	width++;

	height = surface->h-1;
	height = (height >> 1) | height;
	height = (height >> 2) | height;
	height = (height >> 4) | height;
	height = (height >> 8) | height;
	height = (height >> 16) | height;
	height++;

	if(width != surface->w || height != surface->h)
	{
		SDL_Surface * newSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, width,
			height, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000);
		SDL_SetAlpha(surface, 0, 0);
		SDL_BlitSurface(surface, NULL, newSurface, NULL);
		surface = newSurface;
		created_new = TRUE;
	}

	// Determine number of colors in source image.
	numberOfColors = surface->format->BytesPerPixel;
	if(numberOfColors == 4)
		textureFormat = GL_BGRA;
	else if(numberOfColors == 3)
		textureFormat = GL_BGR;
	else
	{
		fprintf(stderr, "%s: Image format not supported.\n", name);
		exit(EXIT_FAILURE);
	}

	// Put the SDL image on the GL texture
	glGenTextures(1, &texture);
	handle_opengl_error(name, "glGenTextures");
	glBindTexture(GL_TEXTURE_2D, texture);
	handle_opengl_error(name, "glBindTexture");
	glTexImage2D(GL_TEXTURE_2D, 0, numberOfColors, surface->w, surface->h,
				 0, textureFormat, GL_UNSIGNED_BYTE, surface->pixels);
	handle_opengl_error(name, "glTexImage2D");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	handle_opengl_error(name, "glTexParameteri");
	glBindTexture(GL_TEXTURE_2D, 0);

	if(created_new)
		SDL_FreeSurface(surface);
	return texture;
}

static void resources_free_surface_cb(void * pointer)
{
	SDL_Surface * surface = (SDL_Surface*)pointer;
	assert(pointer != NULL);
	SDL_FreeSurface(surface);
}

static void resources_free_texture_cb(void * pointer)
{
	GLuint tex = (GLuint)pointer;
	glDeleteTextures(1, &tex);
}

static void resources_free_spriteinfo_cb(void * pointer)
{
	SpriteInfo * spriteinfo = (SpriteInfo*)pointer;
	sprite_destroy_info(spriteinfo);
}

static int resources_find_vboinfo_cb(void * data, void * userdata)
{
	VboInfo * info = (VboInfo*)data;
	GLuint * handle = (GLuint*)userdata;
	if(&info->handle == handle)
		return 0;
	return -1;
}

static void resources_reload_textures()
{
	SDict * iter = m_textures;
	
	m_surfaces = sdict_free_full(m_surfaces, &resources_free_surface_cb);
	
	while(iter != NULL)
	{
		iter->value = (void*)resources_load_gl_image(iter->key);
		iter = iter->next;
	}
}

static void resources_reload_vbos()
{
	SList * iter = m_vbos;
	while(iter != NULL)
	{
		VboInfo * info = (VboInfo*)iter->data;

		// Create new handle
		glGenBuffers(1, &info->handle);

		// Copy data to video memory
		glBindBuffer(GL_ARRAY_BUFFER, info->handle);
		glBufferData(GL_ARRAY_BUFFER, info->size, info->data, GL_STATIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		
		iter = iter->next;
	}
}

//##############################################################################
// Exported functions
//##############################################################################
void resources_initialize()
{
	int initted = 0;

	// Compare linktime and compiletime versions of SDL_image
	SDL_version compile_version;
	const SDL_version * link_version = IMG_Linked_Version();
	SDL_IMAGE_VERSION(&compile_version);
	if(compile_version.major != link_version->major ||
		compile_version.minor != link_version->minor ||
		compile_version.patch != link_version->patch)
	{
		fprintf(stdout, "Compiled with SDL_image version: %d.%d.%d\n",
			compile_version.major,
			compile_version.minor,
			compile_version.patch);
		fprintf(stdout, "Running with SDL_image version: %d.%d.%d\n",
			link_version->major,
			link_version->minor,
			link_version->patch);
	}

	// Initialize PNG support
	initted = IMG_Init(IMG_INIT_PNG);
	if((initted & IMG_INIT_PNG) != IMG_INIT_PNG)
	{
		fprintf(stderr, "IMG_Init: Failed to initialize png support!\n");
		fprintf(stderr, "%s\n", IMG_GetError());
		exit(EXIT_FAILURE);
	}
}

void resources_reload()
{
	resources_reload_textures();
	resources_reload_vbos();
}

void resources_unload()
{
	m_surfaces = sdict_free_full(m_surfaces, &resources_free_surface_cb);
	m_textures = sdict_free_full(m_textures, &resources_free_texture_cb);
	m_sprites = sdict_free_full(m_sprites, &resources_free_spriteinfo_cb);

	IMG_Quit();
}

GLuint * resources_create_vbo(GLsizei size, void * data)
{
	// Initialize VboInfo structure
	VboInfo * info = (VboInfo*)siphon_malloc(sizeof(VboInfo));
	info->data = data;
	info->size = size;
	glGenBuffers(1, &info->handle);

	// Copy data to video memory
	glBindBuffer(GL_ARRAY_BUFFER, info->handle);
	glBufferData(GL_ARRAY_BUFFER, info->size, info->data, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	// Store VboInfo structure
	m_vbos = slist_add(m_vbos, (void*)info);

	return &info->handle;
}

void resources_destroy_vbo(GLuint * handle)
{
	// Find the VboInfo structure that contains the handle pointed to
	SList * node = slist_find(m_vbos, &resources_find_vboinfo_cb, handle);
	assert(node != NULL);
	if(node != NULL)
	{
		VboInfo * info = (VboInfo*)node->data;
		glDeleteBuffers(1, &info->handle);
		m_vbos = slist_remove(m_vbos, node->data);
		siphon_free(info);
	}
}

GLuint * resources_get_gl_image(const char * name)
{
	SDict * node = sdict_find(m_textures, name);
	if(node != NULL)
	{
		return (GLuint*)&(node->value);
	}
	else
	{
		GLuint image = resources_load_gl_image(name);
		m_textures = sdict_add(m_textures, name, (void*)image);
		return (GLuint*)&(m_textures->value);
	}
}

SDL_Surface * resources_get_image(const char * name)
{
	SDict * node = sdict_find(m_surfaces, name);
	if(node != NULL)
	{
		return (SDL_Surface*)node->value;
	}
	else
	{
		SDL_Surface * surface = resources_load_image(name);
		m_surfaces = sdict_add(m_surfaces, name, (void*)surface);
		return (SDL_Surface*)m_surfaces->value;
	}
}

SpriteInfo * resources_get_spriteinfo(const char * name)
{
	SDict * node = sdict_find(m_sprites, name);
	if(node != NULL)
	{
		return (SpriteInfo*)node->value;
	}
	else
	{
		SpriteInfo * spriteinfo = resources_load_spriteinfo(name);
		m_sprites = sdict_add(m_sprites, name, (void*)spriteinfo);
		return (SpriteInfo*)m_sprites->value;
	}
}

#if defined(__linux__)
#include <dirent.h>
DirectoryList * resources_create_list(const char * name, ListFilterFunc filter)
{
	DIR * directory = NULL;
	struct dirent * entry = NULL;
	DirectoryList * list = NULL;
	unsigned int capacity = DIRECTORY_LIST_DEFAULT_CAPACITY;

	assert(name != NULL);

	directory = opendir(name);
	if(directory == NULL)
	{
		fprintf(stderr, "Failed to list files in: %s", name);
		perror("");
		exit(EXIT_FAILURE);
	}

	list = (DirectoryList*)siphon_malloc(sizeof(DirectoryList));
	list->count = 0;
	list->items = (char**)siphon_malloc(capacity * sizeof(char**));

	while((entry = readdir(directory)) != NULL)
	{
		if(filter && !filter(entry->d_name))
			continue;

		list->count++;
		if(list->count > capacity)
		{
			capacity *= 2;
			list->items = (char**)siphon_realloc(list->items,
				capacity * sizeof(char**));
		}

		list->items[list->count-1] = siphon_strdup(entry->d_name);
	}

	closedir(directory);

	return list;
}
#endif

#if defined(_WIN32)
#include <io.h>
DirectoryList * resources_create_list(const char * name, ListFilterFunc filter)
{
	int handle = -1;
	struct _finddata_t entry;
	DirectoryList * list = NULL;
	char buffer[255];
	unsigned int capacity = DIRECTORY_LIST_DEFAULT_CAPACITY;

	assert(name != NULL);
	sprintf(buffer, "%s/*", name);
	handle = _findfirst(buffer, &entry);
	if(handle < 0)
	{
		fprintf(stderr, "Failed to list files in: %s", name);
		perror("");
		exit(EXIT_FAILURE);
	}

	list = (DirectoryList*)siphon_malloc(sizeof(DirectoryList));
	list->count = 0;
	list->items = (char**)siphon_malloc(capacity * sizeof(char**));

	do
	{
		if(filter && !filter(entry.name))
			continue;

		list->count++;
		if(list->count > capacity)
		{
			capacity *= 2;
			list->items = (char**)siphon_realloc(list->items,
				capacity * sizeof(char**));
		}

		list->items[list->count-1] = siphon_strdup(entry.name);

	} while(_findnext(handle, &entry) != -1);

	_findclose(handle);

	return list;
}
#endif

void resources_destroy_list(DirectoryList * list)
{
	int i = 0;

	assert(list != NULL);

	for(i = 0; i < list->count; i++)
	{
		siphon_free(list->items[i]);
	}

	siphon_free(list->items);
	siphon_free(list);
}
