//##############################################################################
// siphon
//
// sound.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <SDL_mixer.h>
#include <assert.h>

#include "sound.h"
#include "common.h"
#include "config.h"

//##############################################################################
// Static variables
//##############################################################################
static Mix_Chunk * m_chunks[SOUND_SAMPLE_MAX];
static Mix_Music * m_current_music = NULL;
static bool m_sound_enabled = FALSE;
static bool m_mp3_enabled = FALSE;

//##############################################################################
// Exported functions
//##############################################################################
void sound_initialize()
{
	// Compare linktime and compiletime versions of SDL_Mixer
	SDL_version compile_version;
	const SDL_version * link_version = Mix_Linked_Version();
	SDL_MIXER_VERSION(&compile_version);
	if(compile_version.major != link_version->major ||
		compile_version.minor != link_version->minor ||
		compile_version.patch != link_version->patch)
	{
		fprintf(stdout, "Compiled with SDL_mixer version: %d.%d.%d\n",
			compile_version.major,
			compile_version.minor,
			compile_version.patch);
		fprintf(stdout, "Running with SDL_mixer version: %d.%d.%d\n",
			link_version->major,
			link_version->minor,
			link_version->patch);
	}

	// Initialize SDL_Mixer with mp3 support
	if(Mix_Init(MIX_INIT_MP3) & MIX_INIT_MP3)
	{
		m_mp3_enabled = TRUE;
	}
	else
	{
		fprintf(stderr, "Failed to initialize MP3 playback\n");
		return;
	}

	// Open sound mixer
	if(Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 1024) == -1)
	{
		fprintf(stderr, "Mix_OpenAudio: %s\n", Mix_GetError());
		return;
	}

	m_sound_enabled = TRUE;

	// Load chunks
	memset((void*)m_chunks, 0x0, sizeof(m_chunks));
	m_chunks[SOUND_FLIP_SWITCH] = Mix_LoadWAV("res/sound_flip_switch.wav");
	m_chunks[SOUND_EXPLODE] = Mix_LoadWAV("res/sound_explode.wav");
	m_chunks[SOUND_SHOOT] = Mix_LoadWAV("res/sound_shoot.wav");
	m_chunks[SOUND_THUMP] = Mix_LoadWAV("res/sound_thump.wav");
	m_chunks[SOUND_JUMP] = Mix_LoadWAV("res/sound_jump.wav");

	// Adjust volume of THUMP chunk
	if(m_chunks[SOUND_THUMP] != NULL)
		Mix_VolumeChunk(m_chunks[SOUND_THUMP], MIX_MAX_VOLUME / 3);
	if(m_chunks[SOUND_JUMP] != NULL)
		Mix_VolumeChunk(m_chunks[SOUND_JUMP], MIX_MAX_VOLUME / 3);
}

void sound_cleanup()
{
	int i = 0;

	if(!m_sound_enabled)
		return;

	for(i = 0; i < SOUND_SAMPLE_MAX; i++)
	{
		if(m_chunks[i] != NULL)
		{
			Mix_FreeChunk(m_chunks[i]);
			m_chunks[i] = NULL;
		}
	}

	if(m_current_music != NULL)
		Mix_FreeMusic(m_current_music);

	// Close sound mixer
	Mix_CloseAudio();

	// Cleanup SDL_Mixer (unloads dynamically loaded libraries)
	if(m_mp3_enabled)
		Mix_Quit();
}

void sound_play_music(const char * music)
{
	if(!m_sound_enabled || !m_mp3_enabled || !config.music_enabled)
		return;

	if(m_current_music != NULL)
		Mix_FreeMusic(m_current_music);
	m_current_music = Mix_LoadMUS(music);

	if(m_current_music != NULL)
	{
		if(Mix_FadeInMusic(m_current_music, -1, 2000) == -1)
		{
			fprintf(stderr, "Mix_FadeInMusic: %s\n", Mix_GetError());
		}
	}
}

void sound_play_sample(SoundSample sample)
{
	int i = (int)sample;

	assert(sample >= 0 && sample < SOUND_SAMPLE_MAX);

	if(!m_sound_enabled || !config.sample_enabled)
		return;

	if(m_chunks[i] != NULL)
	{
		if(Mix_PlayChannel(-1, m_chunks[i], 0) == -1)
		{
			fprintf(stderr, "Mix_PlayChannel: %s\n", Mix_GetError());
		}
	}
}

bool sound_music_is_playing()
{
	return Mix_PlayingMusic();
}

bool sound_music_is_paused()
{
	if(Mix_PlayingMusic() && Mix_PausedMusic())
		return TRUE;
	else
		return FALSE;
}

void sound_pause_music()
{
	Mix_PauseMusic();
}

void sound_stop_music()
{
	Mix_HaltMusic();
}

void sound_resume_music()
{
	if(!config.music_enabled)
		return;
	Mix_ResumeMusic();
}
