//##############################################################################
// siphon
//
// game.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <SDL.h>
#include <assert.h>
#include <math.h>

#include "game.h"
#include "font.h"
#include "main.h"
#include "sound.h"
#include "common.h"
#include "config.h"
#include "transition.h"
#include "opengl_helper.h"
#include "gameobjects/signpost.h"
#include "gameobjects/player.h"
#include "states/levelselect.h"
#include "fixedpoint.h"
#include "gameobject.h"
#include "bresenham.h"
#include "viewport.h"
#include "map.h"

//##############################################################################
// Static variables
//##############################################################################
static GameStatus m_status = GAMESTATUS_NONE;
static Map * m_map = NULL;
static Font * m_rpg_font = NULL;
static Font * m_small_font = NULL;
static Font * m_fixed_font = NULL;
static GameObject * m_focus = NULL;
static GameObjectNode * m_gameobjects = NULL;
static Uint8 * m_empty_keystate = NULL;
static Vector m_root = { 0, 0 };
static char * m_message = NULL;
static int m_message_timeleft = 0;

//##############################################################################
// Camera
//##############################################################################
struct
{
	FixedVector pos;
	FixedVector size;
	int target_offset;
	int offset;
} m_camera;

//##############################################################################
// Static functions
//##############################################################################
static void game_signpost_helper(GameObject * go_focus, GameObject * other)
{
	assert(go_focus);
	assert(other);

	if(other->type & SIGNPOST)
	{
		game_set_message(signpost_get_message(other));
	}
}

static GameObject * game_get_next_focus()
{
	GameObjectNode * node = m_gameobjects;
	while(node)
	{
		if(node->go == m_focus)
		{
			node = node->next;
			while(node)
			{
				if(node->go->type & PLAYER)
				{
					Player * player = (Player*)node->go;
					if(!player->is_dead)
						return node->go;
				}
				node = node->next;
			}
			node = m_gameobjects;
			while(node)
			{
				if(node->go->type & PLAYER)
				{
					Player * player = (Player*)node->go;
					if(!player->is_dead)
						return node->go;
				}
				node = node->next;
			}
			break;
		}
		node = node->next;
	}
	return NULL;
}

static int collides(const GameObject * go, int x, int y)
{
	Vector tl, br;
	br.x = (x + go->size.x - 1) >> m_map->tileshift;
	br.y = (y + go->size.y - 1) >> m_map->tileshift;

	for(tl.x = x >> m_map->tileshift; tl.x <= br.x; tl.x++)
	{
		for(tl.y = y >> m_map->tileshift; tl.y <= br.y; tl.y++)
		{
			int tilenum = map_get_tilenum(m_map, tl);
			if(tilenum == -1 || (tilenum >= 4))
				return 1;
		}
	}
	return 0;
}

static BresenhamAction bresenham_helper(FIXED_POINT x, FIXED_POINT y, void * data)
{
	GameObject * go = (GameObject*)data;

	assert(go->movement_count < sizeof(go->movement) / sizeof(FixedVector));
	if(go->movement_count >= sizeof(go->movement) / sizeof(FixedVector))
		return STOP;
	go->movement[go->movement_count] = go->pos;

	if(!collides(go, TOINT(x), TOINT(y)))
	{
		go->pos.x = x;
		go->pos.y = y;
		go->movement_count++;
		return NEXT;
	}
	else if(!collides(go, TOINT(go->pos.x), TOINT(y)) && go->pos.y != y)
	{
		go->pos.y = y;
		go->movement_count++;
		return REVERT_X;
	}
	else if(!collides(go, TOINT(x), TOINT(go->pos.y)) && go->pos.x != x)
	{
		go->pos.x = x;
		go->movement_count++;
		return REVERT_Y;
	}
	return STOP;
}

static void game_handle_collision(GameObject * go)
{
	int retval = 0;
	go->movement_count = 0;
	go->vel = go->req_vel;

	retval = bresenham_line_do(go->pos.x, go->pos.y, go->req_pos.x,
		go->req_pos.y, &bresenham_helper, (void*)go);

	if(retval == 1)
		return;

	if(bresenham_helper(go->req_pos.x, go->req_pos.y, (void*)go) == NEXT)
	{
		go->pos.x = go->req_pos.x;
		go->pos.y = go->req_pos.y;
	}
}

static void game_draw_hud()
{
	Vector fpos, textsize;
	char buffer[16];

	int seconds = m_map->tickcount / 60;
	int minutes = seconds / 60;

	sprintf(buffer, "%02d:%02d", minutes, seconds % 60);

	fpos = viewport_get_size();
	textsize = font_measure_text(m_rpg_font, buffer);

	fpos.y = 5;
	fpos.x -= textsize.x + 5;

	font_draw_text(m_rpg_font, fpos, buffer);

	sprintf(buffer, "%d/%d", m_map->finished_players, m_map->total_players);
	textsize = font_measure_text(m_rpg_font, buffer);

	fpos.x = 5;
	fpos.y = 5;
	font_draw_text(m_rpg_font, fpos, buffer);

	if(m_message_timeleft > 0)
	{
		int top = 0;
		Vector topleft, size;
		fpos = viewport_get_size();
		m_message_timeleft--;
		textsize = font_measure_text(m_fixed_font, m_message);

		top = fpos.y - m_fixed_font->height * 3 - 4;

		glColor4f(1.0f, 1.0f, 1.0f, 0.3f);
		glBegin(GL_QUADS);
		glVertex2f(0.0f, top);
		glVertex2f(fpos.x,  top);
		glVertex2f(fpos.x, fpos.y);
		glVertex2f(0, fpos.y);
		glEnd();

		glColor4f(1.0f, 1.0f, 1.0f, 1.5f);
		topleft.x = 2;
		topleft.y = top + 2;

		size.x = fpos.x - 4;
		size.y = m_fixed_font->height * 3;

		font_draw_dialog(m_fixed_font, topleft, size, m_message);
	}
}

static bool colliding(GameObject * a, GameObject * b)
{
	if(a->pos.y 					> b->pos.y + TOFIX(b->size.y)) return FALSE;
	if(a->pos.x + TOFIX(a->size.x)	< b->pos.x) return FALSE;
	if(a->pos.y + TOFIX(a->size.y)	< b->pos.y) return FALSE;
	if(a->pos.x 					> b->pos.x + TOFIX(b->size.x)) return FALSE;
	return TRUE;
}

static void add_gameobject(GameObject * gameobject)
{
	GameObjectNode * node = NULL;

	assert(gameobject);

	if(gameobject->type & PLAYER)
	{
		m_map->total_players++;
		m_focus = gameobject;
	}

	node = gameobject_create_node(gameobject);

	if(m_gameobjects == NULL)
		m_gameobjects = node;
	else
		gameobject_append_node(m_gameobjects, node);
}

static void game_handle_event(SDL_Event * event)
{
	if(event->type != SDL_KEYDOWN)
		return;

	switch(event->key.keysym.sym)
	{
		case(SDLK_ESCAPE):
			sound_stop_music();
			m_status = GAMESTATUS_CANCELED;
			transition_from_to("game", "levelselect", TRANSITION_SLIDE_RIGHT);
			break;
		case(SDLK_p):
			m_status = GAMESTATUS_PAUSED;
			main_set_gamestate("paused");
			break;
		case(SDLK_n):
			m_focus = game_get_next_focus();
			break;
		case(SDLK_m):
			config.music_enabled = !config.music_enabled;
			if(config.music_enabled)
			{
				if(sound_music_is_paused())
					sound_resume_music();
				else
					sound_play_music(properties_get_value(m_map->properties, "music"));
			}
			else
				sound_pause_music();
		default:
			break;
	}

}

static void game_enter_state()
{
	assert(m_map != NULL);
	m_status = GAMESTATUS_NONE;
	if(sound_music_is_paused())
	{
		sound_resume_music();
	}
	else if(!sound_music_is_playing())
	{
		sound_play_music(properties_get_value(m_map->properties, "music"));
	}
}

static void game_leave_state(const char * next_state)
{
	if(next_state == NULL)
		return;

	assert(m_status != GAMESTATUS_NONE);

	// Dont pause music when going into the paused state
	if(strcmp(next_state, "paused") != 0)
		sound_pause_music();
}

static void game_initialize()
{
	int numkeys;

	m_camera.pos.x = TOFIX(0);
	m_camera.pos.y = TOFIX(0);
	m_camera.size.x = 24;
	m_camera.size.y = 64;

	if(m_rpg_font == NULL)
		m_rpg_font = font_create("res/rpgfont.png", " abcdefghijklmnopqrstuvwxy"
			"zABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`'*#=[]\"");

	if(m_small_font == NULL)
		m_small_font = font_create("res/small_font.png", " abcdefghijklmnopqrst"
			"uvwxyz-1234567890");

	if(m_fixed_font == NULL)
		m_fixed_font = font_create("res/fixedfont.png", " abcdefghijklmnopqrstu"
			"vwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-.");

	SDL_GetKeyState(&numkeys);
	m_empty_keystate = (Uint8*)siphon_malloc(sizeof(Uint8) * numkeys);
	memset(m_empty_keystate, 0, sizeof(Uint8) * numkeys);
}

static void game_draw(void)
{
	int map_x, map_y;
	int vertbars = 0;
	int horzbars = 0;
	Vector viewport_size = viewport_get_size();
	GameObjectNode * cur = NULL;

	assert(m_focus != NULL);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	if(m_camera.pos.x > m_focus->pos.x)
	{
		m_camera.pos.x = m_focus->pos.x;
		m_camera.target_offset -= 3;
		m_camera.target_offset = MAX(m_camera.target_offset, -15);
	}
	else if(m_camera.pos.x + TOFIX(m_camera.size.x) <
		m_focus->pos.x + TOFIX(m_focus->size.x))
	{
		m_camera.pos.x = m_focus->pos.x +
			TOFIX(m_focus->size.x) - TOFIX(m_camera.size.x);
		m_camera.target_offset += 3;
		m_camera.target_offset = MIN(m_camera.target_offset, 15);
	}

	if(m_camera.pos.y > m_focus->pos.y)
	{
		m_camera.pos.y = m_focus->pos.y;
	}
	else if(m_camera.pos.y + TOFIX(m_camera.size.y) <
		m_focus->pos.y + TOFIX(m_focus->size.y))
	{
		m_camera.pos.y = m_focus->pos.y +
			TOFIX(m_focus->size.y) - TOFIX(m_camera.size.y);
	}

	map_x = TOINT(m_camera.pos.x) - (viewport_size.x / 2) +
		(m_camera.size.x / 2) + m_camera.offset;
	map_y = TOINT(m_camera.pos.y) - (viewport_size.y / 2) +
		(m_camera.size.y / 2);

	map_x = MAX(map_x, 0);
	map_y = MAX(map_y, 0);
	map_x = MIN(map_x, (int)(m_map->w * m_map->tilesize - viewport_size.x));
	map_y = MIN(map_y, (int)(m_map->h * m_map->tilesize - viewport_size.y));


	if((unsigned int)viewport_size.x > m_map->w * m_map->tilesize)
	{
		map_x = -(viewport_size.x / 2) + (m_map->w * m_map->tilesize / 2);
		vertbars = (viewport_size.x / 2) - (m_map->w * m_map->tilesize / 2);
	}
	if((unsigned int)viewport_size.y > m_map->h * m_map->tilesize)
	{
		map_y = -(viewport_size.y / 2) + (m_map->h * m_map->tilesize / 2);
		horzbars = (viewport_size.y / 2) - (m_map->h * m_map->tilesize / 2);
	}

	m_root.x = map_x;
	m_root.y = map_y;
	if(vertbars || horzbars)
		viewport_scissor_start(vertbars, horzbars,
			viewport_size.x - 2 * vertbars,
			viewport_size.y - 2 * horzbars);

	if(config.flip_map_horizontally)
	{
		glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
		glTranslatef(-viewport_size.x, -viewport_size.y, 0.0f);
	}
	map_draw(m_map, m_root);

	if(vertbars || horzbars)
		viewport_scissor_stop();


	cur = m_gameobjects;
	while(cur != NULL)
	{
		assert(cur->go->draw != NULL);
		cur->go->draw(cur->go, map_x, map_y);
		cur = cur->next;
	}

	if(config.flip_map_horizontally)
	{
		glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
		glTranslatef(-viewport_size.x, -viewport_size.y, 0.0f);
	}

	game_draw_hud();
}

static void game_logic()
{
	GameObjectNode * cur = NULL;
	if(SDL_GetKeyState(NULL)[SDLK_r])
		m_map->tickcount = MAX(0, m_map->tickcount-1);
	else
		m_map->tickcount++;

	if(m_camera.offset < m_camera.target_offset)
	{
		int diff = m_camera.target_offset - m_camera.offset;
		m_camera.offset += MIN(diff, 1);
	}
	else if(m_camera.offset > m_camera.target_offset)
	{
		int diff = m_camera.offset - m_camera.target_offset;
		m_camera.offset -= MIN(diff, 1);
	}

	cur = m_gameobjects;
	m_map->alive_players = 0;
	m_map->finished_players = 0;
	while(cur != NULL)
	{
		if(cur->go->type & PLAYER)
		{
			Player * player = (Player*)cur->go;
			if(!player->is_dead)
				m_map->alive_players++;
			if(player->has_finished)
				m_map->finished_players++;
		}

		if(!cur->go->logic_updated)
		{
			assert(cur->go->logic != NULL);
			cur->go->logic(cur->go);
			cur->go->logic_updated = TRUE;
		}
		cur = cur->next;
	}

	cur = m_gameobjects;
	while(cur != NULL)
	{
		if(SDL_GetKeyState(NULL)[SDLK_r])
		{
			game_handle_collision(cur->go);
		}
		else
		{
			if(cur->go->type & PLAYER || cur->go->type & ROPE)
				gameobject_push_history(cur->go);
			game_handle_collision(cur->go);
		}
		cur->go->logic_updated = FALSE;
		cur = cur->next;
	}

	game_do_colliders(m_focus, &game_signpost_helper);

	map_logic(m_map);

	if(m_map->alive_players == 0)
	{
		sound_stop_music();
		m_status = GAMESTATUS_DIED;
		transition_from_to("game", "levelselect", TRANSITION_SLIDE_RIGHT);
	}

	if(m_map->finished_players == m_map->total_players)
	{
		levelselect_increase_progress();
		sound_stop_music();
		m_status = GAMESTATUS_COMPLETED_LEVEL;
		transition_from_to("game", "levelselect", TRANSITION_SLIDE_RIGHT);
	}
}

static void game_cleanup()
{
	siphon_free(m_empty_keystate);
	gameobject_destroy(m_gameobjects);
	if(m_rpg_font)
		font_destroy(m_rpg_font);
	if(m_small_font)
		font_destroy(m_small_font);
	if(m_fixed_font)
		font_destroy(m_fixed_font);
	if(m_map != NULL)
		map_destroy(m_map);
	if(m_message)
		siphon_free(m_message);
}

//##############################################################################
// Exported functions
//##############################################################################
void game_load_map(const char * name)
{
	GameObjectPosition * pos = NULL;

	m_message_timeleft = 0;

	config.player_can_jet = FALSE;
	config.player_can_climb = FALSE;
	config.player_can_walljump = FALSE;
	config.flip_gravity = FALSE;
	config.flip_left_right = FALSE;
	config.flip_map_horizontally = FALSE;
	config.flip_water_land = FALSE;
	config.flip_rope = FALSE;

	if(m_map != NULL)
		map_destroy(m_map);
	m_map = map_create(name);

	gameobject_destroy(m_gameobjects);
	m_gameobjects = NULL;

	pos = m_map->gameobjects;
	while(pos)
	{
		CreateArgs args;
		memset(&args, 0x0, sizeof(args));
		args.addfunc = &add_gameobject;
		args.properties = pos->properties;
		args.map = m_map;
		registry_create(pos->name, &args);

		pos = pos->next;
	}
}

GameStatus game_get_status()
{
	return m_status;
}

Uint8 * game_get_keystate(GameObject * gameobject)
{
	if(gameobject == m_focus)
		return SDL_GetKeyState(NULL);
	else
		return m_empty_keystate;
}

bool game_is_on_screen(GameObject * gameobject)
{
	Vector size = viewport_get_size();
	Vector root = m_root;

	assert(gameobject != NULL);

	root.x -= 50; // a gameobject is considered on screen if it's in the
	root.y -= 50; // visible area of the world, or in the 50 pixels wide border
	size.x += 100; // around that area.
	size.y += 100;

	if(root.x > TOINT(gameobject->pos.x) + gameobject->size.x)
		return FALSE;
	if(root.y > TOINT(gameobject->pos.y) + gameobject->size.y)
		return FALSE;
	if(root.x + size.x < TOINT(gameobject->pos.x))
		return FALSE;
	if(root.y + size.y < TOINT(gameobject->pos.y))
		return FALSE;

	return TRUE;
}

void game_do_colliders(GameObject * gameobject, collider_callback cb)
{
	GameObjectNode * node = NULL;
	for(node = m_gameobjects; node; node = node->next)
	{
		if(node->go == gameobject)
			continue;
		if(colliding(node->go, gameobject))
			cb(gameobject, node->go);
	}
}

void game_set_message(const char * text)
{
	m_message_timeleft = 100;
	if(m_message != NULL && strcmp(text, m_message) == 0)
		return;
	if(m_message != NULL)
		siphon_free(m_message);
	m_message = siphon_strdup(text);
}

//##############################################################################
// State registration
//##############################################################################
CONSTRUCTOR(game_const)
{
	StateInterface interface;
	memset(&interface, 0x0, sizeof(StateInterface));
	interface.name = "game";
	interface.initialize = &game_initialize;
	interface.cleanup = &game_cleanup;
	interface.logic = &game_logic;
	interface.draw = &game_draw;
	interface.handle_event = &game_handle_event;
	interface.enter_state = &game_enter_state;
	interface.leave_state = &game_leave_state;

	main_register_state(interface);
}
