//##############################################################################
// siphon
//
// levelselect.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include "levelselect.h"
#include "main.h"
#include "font.h"
#include "vector.h"
#include "viewport.h"
#include "transition.h"
#include "common.h"
#include "game.h"
#include "config.h"
#include "resources.h"

//##############################################################################
// Defines
//##############################################################################
#define MAPS_DIRECTORY "res/maps"

//##############################################################################
// Type definitions
//##############################################################################
typedef struct
{
	char * nicename;
	char * filename;
} Level;

//##############################################################################
// Static variables
//##############################################################################
static Font * m_rpg_font = NULL;
static Font * m_small_font = NULL;
static Level m_levels[255];
static int m_offset = 0;
static int m_level_count = 0;
static int m_current_level = 0;
static DirectoryList * m_directory_list = NULL;

//##############################################################################
// Static functions
//##############################################################################
static void levelselect_draw_background()
{
	Vector size = viewport_get_size();

	glBegin(GL_QUADS);
		glColor4f(0.0f, 0.15f, 0.47f, 1.0f);
		glVertex2f(0, 0);
		glVertex2f(size.x, 0);
		glColor4f(0.39f, 0.76f, 1.0f, 1.0f);
		glVertex2f(size.x, size.y);
		glVertex2f(0, size.y);
	glEnd();

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
}

static bool filter_maps(const char * name)
{
	int i = 0;
	const char * map_str = "map";

	while(name[i] && map_str[i])
	{
		if(name[i] != map_str[i])
			return FALSE;
		i++;
	}
	return map_str[i] == '\0';
}

static int sort_maps(const void * l, const void * r)
{
	return strcmp(*(char**)l, *(char**)r);
}

static void levelselect_initialize()
{
	int i = 0;

	m_level_count = 0;
	m_current_level = 0;

	m_directory_list = resources_create_list(MAPS_DIRECTORY, filter_maps);

	qsort(m_directory_list->items, m_directory_list->count, sizeof(char*),
		sort_maps);

	for(i = 0; i < m_directory_list->count; i++)
	{
		char * nicename = strchr(m_directory_list->items[i], '_');
		if(nicename != NULL && *(nicename + 1))
			m_levels[m_level_count].nicename = nicename + 1;
		else
			m_levels[m_level_count].nicename = m_directory_list->items[i];
		m_levels[m_level_count++].filename = m_directory_list->items[i];
	}

	m_levels[m_level_count].nicename = NULL;
	m_levels[m_level_count].filename = NULL;

	if(m_rpg_font == NULL)
		m_rpg_font = font_create("res/rpgfont.png", " abcdefghijklmnopqrstuvwxy"
			"zABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`'*#=[]\"");
	if(m_small_font == NULL)
		m_small_font = font_create("res/small_font.png", " abcdefghijklmnopqrst"
			"uvwxyz-1234567890");
}

static void levelselect_enter_state()
{
}

static void levelselect_handle_event(SDL_Event * event)
{
	switch(event->type)
	{
		case SDL_KEYDOWN:
			if(event->key.keysym.sym == SDLK_RETURN)
			{
				if(m_current_level <= config.progress)
				{
					char buffer[255];
					sprintf(buffer, "%s/%s", MAPS_DIRECTORY,
						m_levels[m_current_level].filename);
					game_load_map(buffer);
					if(config.progress >= m_level_count)
						config.player_can_jet = TRUE;

					transition_from_to("levelselect", "game",
						TRANSITION_FADE_BLACK);
				}
			}
			else if(event->key.keysym.sym == SDLK_ESCAPE)
			{
				transition_from_to("levelselect", "intro",
					TRANSITION_SLIDE_RIGHT);
			}
			else if(event->key.keysym.sym == SDLK_DOWN)
				m_current_level = (m_current_level+1) % m_level_count;
			else if(event->key.keysym.sym == SDLK_UP)
				m_current_level = (m_current_level+m_level_count-1) %
					m_level_count;
			break;
	}
}

static void levelselect_draw(void)
{
	int i = 0;
	Vector pos;
	Vector textsize;
	Vector size = viewport_get_size();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	levelselect_draw_background();

	textsize = font_measure_text(m_rpg_font, "[Select a level]");

	pos.y = 3;
	pos.x = size.x / 2 - textsize.x / 2;

	glColor4f(1.0f, 1.0f, 1.0f, 0.3f);
	glBegin(GL_QUADS);
	glVertex2f(0, 0);
	glVertex2f(size.x, 0);
	glVertex2f(size.x, 23);
	glVertex2f(0, 23);
	glEnd();
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	font_draw_text(m_rpg_font, pos, "[Select a level]");

	if(config.progress >= m_level_count)
	{
		textsize = font_measure_text(m_small_font, "jetpack enabled");
		pos.y = 18;
		pos.x = size.x / 2 - textsize.x / 2;
		glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
		font_draw_text(m_small_font, pos, "jetpack enabled");
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}

	pos.x = 20;
	pos.y = 23 - m_offset;

	viewport_scissor_start(0, 23, size.x, size.y - 23);
	while(m_levels[i].nicename != NULL)
	{
		char buffer[255];

		sprintf(buffer, "%02d. %s", i+1, m_levels[i].nicename);

		textsize = font_measure_text(m_rpg_font, buffer);

		if(m_current_level == i)
		{
			glColor4f(0.0f, 0.0f, 0.0f, 0.3f);
			glBegin(GL_QUADS);
			glVertex2f(0, pos.y);
			glVertex2f(size.x, pos.y);
			glVertex2f(size.x, pos.y + textsize.y);
			glVertex2f(0, pos.y + textsize.y);
			glEnd();
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		}
		if(i > config.progress)
			glColor4f(0.5f, 0.5f, 0.5f, 1.0f);

		font_draw_text(m_rpg_font, pos, buffer);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

		pos.y += m_rpg_font->height;
		i++;
	}
	viewport_scissor_stop();
}

static void levelselect_logic()
{
	Vector size = viewport_get_size();
	int rows = (size.y - 23) / m_rpg_font->height;
	int diff, step;
	int goal = (m_current_level - (rows / 2)) * m_rpg_font->height;

	goal = MIN((m_level_count-rows) * m_rpg_font->height, goal);
	goal = MAX(0, goal);

	diff = m_offset - goal;

	step = diff / m_rpg_font->height;

	if(diff < 0)
		step -= 1;
	if(diff > 0)
		step += 1;

	if(m_offset != goal)
	{
		if(abs(diff) < abs(step))
			m_offset -= diff;
		else
			m_offset -= step;
	}
}

static void levelselect_cleanup()
{
	font_destroy(m_rpg_font);
	font_destroy(m_small_font);
	resources_destroy_list(m_directory_list);
}

static void levelselect_leave_state(const char * next_state)
{
}

//##############################################################################
// Exported functions
//##############################################################################
void levelselect_increase_progress()
{
	if(config.progress < m_current_level + 1)
		config.progress = m_current_level + 1;
}

//##############################################################################
// State registration
//##############################################################################
CONSTRUCTOR(levelselect_const)
{
	StateInterface interface;
	memset(&interface, 0x0, sizeof(StateInterface));
	interface.name = "levelselect";
	interface.initialize = &levelselect_initialize;
	interface.cleanup = &levelselect_cleanup;
	interface.logic = &levelselect_logic;
	interface.draw = &levelselect_draw;
	interface.handle_event = &levelselect_handle_event;
	interface.enter_state = &levelselect_enter_state;
	interface.leave_state = &levelselect_leave_state;

	main_register_state(interface);
}
