//##############################################################################
// siphon
//
// viewport.c
//
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################
#include <SDL.h>
#include <assert.h>

#include "viewport.h"
#include "config.h"
#include "common.h"
#include "opengl_helper.h"
#include "vector.h"
#include "resources.h"

//##############################################################################
// Type definitions
//##############################################################################
typedef struct
{
	Vector topleft;
	Vector size;
} Rect;

//##############################################################################
// Static constants
//##############################################################################
static const char * CAPTION = "Siphon";
static int SCREEN_BPP = 32;
static int SCREEN_FLAGS = SDL_OPENGL | SDL_DOUBLEBUF | SDL_HWSURFACE | SDL_RESIZABLE;

//##############################################################################
// Static variables
//##############################################################################
static Vector m_native_screen_size = { -1, -1 };
static Vector m_actual_vp_size = { -1, -1 };
static Vector m_actual_vp_offset = { -1, -1 };
static Vector m_ortho_size = { -1, -1 };
static Vector m_screen_size = { -1, -1 };
static SDL_Surface * m_screen = NULL;
static int m_scissor_count = 0;
static Rect scissors[16];
static Uint8 m_iconmask[128] =
{
	0x00,0x00,0x00,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xFC,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x1F,0xC0,0x00,
	0x00,0x00,0x00,0x00
};

//##############################################################################
// Static functions
//##############################################################################
static void viewport_initialize_opengl_context()
{
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepth(1.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

static void configure_viewport()
{
	int pixelsize = (int)(m_screen_size.x / 240.0f + 0.5f);
	if(pixelsize < 1)
		pixelsize = 1;

	glPointSize(pixelsize);

	m_ortho_size.x = m_screen_size.x / pixelsize;
	m_ortho_size.y = m_screen_size.y / pixelsize;
	m_actual_vp_size.x = m_screen_size.x - (m_screen_size.x % pixelsize);
	m_actual_vp_size.y = m_screen_size.y - (m_screen_size.y % pixelsize);
	m_actual_vp_offset.x = (m_screen_size.x % pixelsize) / 2;
	m_actual_vp_offset.y = (m_screen_size.y % pixelsize) / 2;
}

static void screen_resize(int w, int h)
{
	w = MAX(w, 1);
	h = MAX(h, 1);

	if(config.fullscreen)
		SCREEN_FLAGS |= SDL_FULLSCREEN;
	else
		SCREEN_FLAGS &= ~SDL_FULLSCREEN;

	if(m_screen == NULL || m_screen->w != w || m_screen->h != h)
	{
		m_screen = SDL_SetVideoMode(w, h, SCREEN_BPP, SCREEN_FLAGS);
		if(m_screen == NULL)
		{
			fprintf(stderr, "Failed to set video mode: %dx%d at %dbpp\n",
				m_screen_size.x, m_screen_size.y, SCREEN_BPP);
			exit(EXIT_FAILURE);
		}
		if(!config.fullscreen)
		{
			config.screen_width = m_screen->w;
			config.screen_height = m_screen->h;
		}
		m_screen_size.x = m_screen->w;
		m_screen_size.y = m_screen->h;
		configure_viewport();
		viewport_initialize_opengl_context();

		#ifdef _WIN32
			resources_reload();
		#endif
	}
}

static void configure_screensize()
{
	if(config.fullscreen)
	{
		m_screen_size.x = m_native_screen_size.x;
		m_screen_size.y = m_native_screen_size.y;
	}
	else if(config.screen_width != -1)
	{
		m_screen_size.x = config.screen_width;
		m_screen_size.y = config.screen_height;
	}
	else
	{
		m_screen_size.x = 240 * 3;
		m_screen_size.y = 160 * 3;
	}
}

//##############################################################################
// Exported functions
//##############################################################################
Vector viewport_get_size()
{
	return m_ortho_size;
}

void viewport_initialize()
{
	const SDL_VideoInfo * info;

	SDL_WM_SetCaption(CAPTION, "");
	SDL_WM_SetIcon(SDL_LoadBMP("res/icon.bmp"), m_iconmask);

	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	info = SDL_GetVideoInfo();
	m_native_screen_size.x = info->current_w;
	m_native_screen_size.y = info->current_h;

	configure_screensize();
	screen_resize(m_screen_size.x, m_screen_size.y);
}

void viewport_begin_draw()
{
	glViewport(m_actual_vp_offset.x, m_actual_vp_offset.y,
		m_actual_vp_size.x, m_actual_vp_size.y);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, m_ortho_size.x, m_ortho_size.y, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void viewport_end_draw()
{
	// Check for OpenGL errors, if there are no errors
	// the GL_NO_ERROR case will swap buffers and return
	// otherwise the program will be terminated
	char * error_string = NULL;
	GLenum error = glGetError();
	switch(error)
	{
		case(GL_NO_ERROR):
			SDL_GL_SwapBuffers();
			return;
		case GL_INVALID_ENUM:
			error_string = "GL_INVALID_ENUM";
			break;
		case GL_INVALID_VALUE:
			error_string = "GL_INVALID_VALUE";
			break;
		case GL_INVALID_OPERATION:
			error_string = "GL_INVALID_OPERATION";
			break;
		case GL_STACK_OVERFLOW:
			error_string = "GL_STACK_OVERFLOW";
			break;
		case GL_STACK_UNDERFLOW:
			error_string = "GL_STACK_UNDERFLOW";
			break;
		case GL_OUT_OF_MEMORY:
			error_string = "GL_OUT_OF_MEMORY";
			break;
	}

	fprintf(stderr, "OpenGL error: %s\n", error_string);
	exit(EXIT_FAILURE);
}

void viewport_handle_event(SDL_Event * event)
{
	int mult = -1;
	int mspf = 0;
	switch(event->type)
	{
		case SDL_VIDEORESIZE:
			screen_resize(event->resize.w, event->resize.h);
			break;
		case SDL_KEYDOWN:
			if(event->key.keysym.sym == SDLK_F1)
				mult = 1;
			else if(event->key.keysym.sym == SDLK_F2)
				mult = 2;
			else if(event->key.keysym.sym == SDLK_F3)
				mult = 3;
			else if(event->key.keysym.sym == SDLK_F4)
				mult = 4;
			else if(event->key.keysym.sym == SDLK_F5)
				mult = 5;
			else if(event->key.keysym.sym == SDLK_F6)
				mult = 6;
			else if(event->key.keysym.sym == SDLK_F7)
				mult = 7;
			else if(event->key.keysym.sym == SDLK_F8)
				mult = 8;
			if(mult != -1)
			{
				config.fullscreen = FALSE;
				m_screen_size.x = 240 * mult;
				m_screen_size.y = 160 * mult;
				screen_resize(m_screen_size.x, m_screen_size.y);
			}
			else if(event->key.keysym.sym == SDLK_F11)
			{
				config.fullscreen = !config.fullscreen;
				configure_screensize();
				screen_resize(m_screen_size.x, m_screen_size.y);
			}

			mspf = config.milliseconds_per_frame;

			if(event->key.keysym.sym == SDLK_F9)
				config.milliseconds_per_frame = MAX(11, mspf-1);
			else if(event->key.keysym.sym == SDLK_F10)
				config.milliseconds_per_frame = MIN(21, mspf+1);
			else if(event->key.keysym.sym == SDLK_F12)
				config.milliseconds_per_frame = 16;
			break;
	}
}

void viewport_scissor_start(int x, int y, int width, int height)
{
	float multx, multy;
	int bottom;

	assert(width >= 0);
	assert(height >= 0);

	m_scissor_count++;

	if(m_scissor_count == 1)
	{
		glEnable(GL_SCISSOR_TEST);
	}
	else
	{
		// Limit the scissor rectangle to be smaller than the rectangle
		// on the top of the scissor stack
		Vector tl = scissors[m_scissor_count-2].topleft;
		Vector sz = scissors[m_scissor_count-2].size;
		if(x < tl.x)
		{
			x = tl.x;
			width -= tl.x - x;
		}
		if(width > sz.x)
			width = sz.x;

		if(y < tl.y)
		{
			y = tl.y;
			height -= tl.y - y;
		}
		if(height > sz.y)
			height = sz.y;
	}

	// Put the new scissor rectangle on the scissor stack
	scissors[m_scissor_count-1].topleft.x = x;
	scissors[m_scissor_count-1].topleft.y = y;
	scissors[m_scissor_count-1].size.x = width;
	scissors[m_scissor_count-1].size.y = height;

	multx = (float)m_screen_size.x / (float)m_ortho_size.x;
	multy = (float)m_screen_size.y / (float)m_ortho_size.y;
	bottom = (int)(m_screen_size.y - y * multy - height * multy);

	glScissor(x * multx, bottom, width * multx, height * multy);
}

void viewport_scissor_stop()
{
	m_scissor_count--;

	if(m_scissor_count == 0)
	{
		glDisable(GL_SCISSOR_TEST);
	}
	else
	{
		// Pop a scissor rectangle from the stack
		Vector tl = scissors[m_scissor_count-1].topleft;
		Vector sz = scissors[m_scissor_count-1].size;
		float multx = (float)m_screen_size.x / (float)m_ortho_size.x;
		float multy = (float)m_screen_size.y / (float)m_ortho_size.y;
		int bottom = (int)(m_screen_size.y - tl.y * multy - sz.y * multy);
		glScissor(tl.x * multx, bottom, sz.x * multx, sz.y * multy);
	}
}

void viewport_cleanup()
{
}
